/*
 * Copyright 2015 by Kappich Systemberatung Aachen
 *
 * This file is part of de.bsvrz.pua.prot.
 *
 * de.bsvrz.pua.prot is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * de.bsvrz.pua.prot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.pua.prot.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Contact Information:
 * Kappich Systemberatung
 * Martin-Luther-Straße 14
 * 52062 Aachen, Germany
 * phone: +49 241 4090 436
 * mail: <info@kappich.de>
 */

package de.bsvrz.pua.prot.client.dataobject;

import de.bsvrz.dav.daf.main.DataState;
import de.bsvrz.pua.prot.util.ProcessingParameter;

/**
 * Datenzustand für PuA-Ergebnisdaten
 *
 * @author Kappich Systemberatung
 */
public enum PuADataState {
	
	/**
	 * Datenzustand für von PuA freigelassene Zellen (z. B. Zellen die nicht aufgefüllt wurden)
	 */
	EMPTY("Leer"),	
	
	/**
	 * Datenzustand für von PuA gebildete Zellen mit einer "Keine-Änderung"-Markierung. Nur wenn PuA zellenweise
	 * "Keine-Änderung"-Markierungen erstellt. ({@link ProcessingParameter#setNoChangeMarker(de.bsvrz.sys.funclib.losb.datk.NoChangeMarker)})
	 */
	NO_CHANGE("Keine Änderung"),
	
	/**
	 * Datenzustand für von PuA gebildete künstliche Datensätze, die nicht über den Datenverteiler empfangen wurden.
	 */
	COMPUTED("Berechnet"),

	/**
	 * Datenzustand für Datensätze, bei denen Fehler bei der Berechnung auftraten
	 */
	ERROR("Fehler"),

	/**
	 * Datenzustand ist aufgrund eines veralteten PuA-Servers nicht verfügbar
	 */
	NOT_AVAILABLE("Nicht verfügbar"),
	
	/**
	 * Datensatztyp für Datensätze die Nutzdaten enthalten (siehe Technische Anforderungen Archivsystem).
	 */
	 DATA(DataState.DATA),

	/**
	 * Datensatztyp für leere Datensätze, die von der Quelle ohne Attributwerte versendet wurden (siehe Technische
	 * Anforderungen Archivsystem).
	 */
	 NO_DATA(DataState.NO_DATA),

	/**
	 * Datensatztyp für leere Datensätze, die vom Datenverteiler generiert wurden, weil keine Quelle für die entsprechenden
	 * Daten existiert. (siehe Technische Anforderungen Archivsystem).
	 */
	 NO_SOURCE(DataState.NO_SOURCE),

	/**
	 * Datensatztyp für leere Datensätze, die vom Datenverteiler generiert wurden, weil nicht die erforderlichen Rechte zum
	 * Empfang der Daten vorliegen. (siehe Technische Anforderungen Archivsystem).
	 */
	 NO_RIGHTS(DataState.NO_RIGHTS),

	/**
	 * Datensatztyp für leere Datensätze, die vom Archivsystem generiert wurden, um eine potentielle Datenlücke zu
	 * markieren. (siehe Technische Anforderungen Archivsystem).
	 */
	 POSSIBLE_GAP(DataState.POSSIBLE_GAP),

	/**
	 * Datensatztyp für leere Datensätze, die vom Archivsystem in jeden Datensatzstrom eingefügt werden, um das Ende eines
	 * Datensatzstroms einer Teilanfrage zu markieren. 
	 * Der Zeitstempel dieses Datensatzes wird PuA-Seitig auf das Ende des Anfrageintervalls gesetzt und entspricht damit 
	 * nicht mehr der Semantik im Archivsystem..
	 */
	END_OF_ARCHIVE(DataState.END_OF_ARCHIVE),
	
	/**
	 * Datensatztyp für leere Datensätze, die vom Archivsystem in den Antwort-Datensatzstrom von Teilanfragen eingefügt
	 * wird, um Bereiche zu markieren, die gelöscht (und nicht gesichert) wurden. Der Datenzeitstempel bei Verwendung
	 * dieses Typs enthält den Datenzeitstempel des ersten gelöschten Datensatz im gelöschten Bereich. (Siehe auch
	 * Technische Anforderungen Archivsystem).
	 */
	 DELETED_BLOCK(DataState.DELETED_BLOCK),

	/**
	 * Datensatztyp für leere Datensätze, die vom Archivsystem in den Antwort-Datensatzstrom von Teilanfragen eingefügt
	 * wird, um Bereiche zu markieren, die ausgelagert (d.h. gesichert und gelöscht) wurden. Der Datenzeitstempel bei
	 * Verwendung dieses Typs enthält den Datenzeitstempel des ersten gelöschten Datensatz im ausgelagerten Bereich. (Siehe
	 * auch Technische Anforderungen Archivsystem).
	 */
	 UNAVAILABLE_BLOCK(DataState.UNAVAILABLE_BLOCK),

	/**
	 * Datensatztyp für leere Datensätze, die vom Datenverteiler versendet werden können, wenn eine Anmeldung von Daten im
	 * Konflikt mit anderen Anmeldungen steht (z.B. mehrere Senken für die gleichen Daten).
	 */
	 INVALID_SUBSCRIPTION(DataState.INVALID_SUBSCRIPTION);

	private final String _name;
	private final int _code;

	PuADataState(DataState dataState) {
		_name = dataState.toString();
		_code = dataState.getCode();
	}

	PuADataState(String name) {
		_name = name;
		_code = 0;
	}

    @Override
	public String toString() {
		return _name;
	}
	
	public static PuADataState of(DataState state){
		for(PuADataState pState : PuADataState.class.getEnumConstants()) {
			if(pState._code == state.getCode()) return pState;
		}
		return PuADataState.NOT_AVAILABLE;
	}
}
