/*
 * Copyright 2010 by Kappich Systemberatung Aachen
 *
 * This file is part of de.bsvrz.pua.prot.
 *
 * de.bsvrz.pua.prot is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * de.bsvrz.pua.prot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.pua.prot.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Contact Information:
 * Kappich Systemberatung
 * Martin-Luther-Straße 14
 * 52062 Aachen, Germany
 * phone: +49 241 4090 436
 * mail: <info@kappich.de>
 */

package de.bsvrz.pua.prot.data;

import de.bsvrz.dav.daf.main.Data;
import de.bsvrz.sys.funclib.losb.util.Tuple;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

/**
 * Bietet eine Kapselung der Protokolldaten in ein Interface, wie es Dav-Daten benutzen. Anders als Dav-Daten sind diese Daten nicht streng an die verwendete
 * Attributgruppe bzw. das verwendete Attribut gebunden (sofern überhaupt eines existiert). Je nach Anwendungsfall bieten sich daher auch Abfragen nach {@link
 * #isDefined()}, {@link #isPlain()} usw. an, auch wenn die Attributgruppe diese Daten eigentlich nicht unterstützen würde. Besonders wichtig ist dies bei
 * Fehlermeldungen von PuA (wie z.B. einer Division durch 0, oder einem "Keine Daten"-Datensatz), da hier die Meldung unabhängig vom Attributtyp über
 * {@link #asTextValue()} abgerufen werden kann.
 *
 * @author Kappich Systemberatung
 * @version $Revision: 0000 $
 */
public abstract class ProtocolData implements Data {

	static final String MESSAGE_NO_CHANGE = "Protokolldaten sind nicht änderbar";

	static final String MESSAGE_NOT_ATOMAR = " ist kein atomarer Wert";

	static final String MESSAGE_NOT_LIST = " ist keine Liste";

	static final String MESSAGE_NOT_ARRAY = " ist kein Array";

	static final String MESSAGE_NOT_ITERABLE = " ist keine Liste und kein Array, daher kann nicht iteriert werden";

	public Data createModifiableCopy() {
		return createUnmodifiableCopy().createModifiableCopy();
	}

	public void setToDefault() {
		throw new UnsupportedOperationException(MESSAGE_NO_CHANGE);
	}

	public TextValue getTextValue(final String itemName) {
		return getItem(itemName).asTextValue();
	}

	public TimeValue getTimeValue(final String itemName) {
		return getItem(itemName).asTimeValue();
	}

	public NumberValue getScaledValue(final String itemName) {
		return getItem(itemName).asScaledValue();
	}

	public NumberValue getUnscaledValue(final String itemName) {
		return getItem(itemName).asUnscaledValue();
	}

	public ReferenceValue getReferenceValue(final String itemName) {
		return getItem(itemName).asReferenceValue();
	}

	public TextArray getTextArray(final String itemName) {
		return getItem(itemName).asTextArray();
	}

	public Array getArray(final String itemName) {
		return getItem(itemName).asArray();
	}

	public TimeArray getTimeArray(final String itemName) {
		return getItem(itemName).asTimeArray();
	}

	public NumberArray getScaledArray(final String itemName) {
		return getItem(itemName).asScaledArray();
	}

	public NumberArray getUnscaledArray(final String itemName) {
		return getItem(itemName).asUnscaledArray();
	}

	public ReferenceArray getReferenceArray(final String itemName) {
		return getItem(itemName).asReferenceArray();
	}

	@Override
	public String toString() {
		return getName() + ": " + valueToString();
	}

	public String valueToString() {
		final StringBuilder builder = new StringBuilder();
		try {
			if(isError()) {
				return "<<Fehler:" + asTextValue().getText() + ">>";
			}
			else if(isPlain() || !isDefined()) {
				return asTextValue().getText();
			}
			final boolean isArray = isArray();
			builder.append(isArray ? "[" : "{");
			for(Iterator<Data> iterator = iterator(); iterator.hasNext();) {
				try {
					final Data item = iterator.next();
					builder.append(isArray ? item.valueToString() : item.toString());
				}
				catch(Exception e) {
					builder.append("<<Fehler:").append(e.getMessage()).append(">>");
				}
				if(iterator.hasNext()) builder.append("; ");
			}
			builder.append(isArray ? "]" : "}");
		}
		catch(Exception e) {
			builder.append("<<").append(e.getMessage()).append(">>");
		}
		return builder.toString();
	}

	/** 
	 * Gibt {@code true} zurück, wenn es sich um einen von PuA generierten Fehlerwert handelt. Dieser kann beispielsweise in einem temporären Attribut
	 * bei einer Division durch 0 entstehen.
	 * @return {@code true}, wenn es sich um einen von PuA generierten Fehlerwert handelt, sonst {@code false}
	 */
	public abstract boolean isError();


	@Override
	public int hashCode() {
		return valueToString().hashCode();
	}

	@Override
	public boolean equals(final Object obj) {
		return obj instanceof Data && equals((Data)obj);
	}

	public boolean equals(final Data other) {
		return valueToString().equals(other.valueToString());
	}

	public Iterator<Data> iterator() {
		if(isPlain()) throw new UnsupportedOperationException(ProtocolData.MESSAGE_NOT_ITERABLE);
		if(isArray()) return new ProtocolDataArrayIterator(asArray());
		return getChildrenList().iterator();
	}

	protected abstract List<Data> getChildrenList();

	/**
	 * Konvertiert die Baumstruktur eines Data-Objektes in eine flache Struktur (Liste), die sich zum Beispiel einfacher in Tabellen darstellen lässt.
	 * @param data Data-Objekt das konvertiert werden soll (wird nicht verändert)
	 * @return flache Kopie des Data-Objektes (Liste von Spaltennamen-Daten-Paaren). Alle enthaltenen Data-Objekte sind Atomar.
	 */
	public static List<Tuple<String, Data>> convertDataToList(final Data data){
		final List<Tuple<String, Data>> result = new ArrayList<>();
		recursiveFlattenData(data, result, "");
		return result;
	}

	private static void recursiveFlattenData(final Data data, final List<Tuple<String, Data>> result, final String prefix) {
		final String myName = prefix + data.getName();
		if(data.isPlain()) {
			result.add(new Tuple<>(myName, data));
		}
		else if(data.isArray()) {
			int i = 0;
			for(Iterator iterator = data.iterator(); iterator.hasNext();) {
				final Data subData = (Data)iterator.next();
				recursiveFlattenData(subData, result, myName + "[" + (i++) + "].");
			}
		}
		else {
			for(Iterator iterator = data.iterator(); iterator.hasNext();) {
				final Data subData = (Data)iterator.next();
				recursiveFlattenData(subData, result, myName + ".");
			}
		}
	}

	/**
	 * Gibt {@code true} zurück, wenn es sich bei diesem Wert um eine NoChange-Markierung handelt. NoChange-Markierungen dieser Art
	 * werden PuA-Seitig generiert, wenn es sich um ein Zustandsprotokoll handelt und die NoChange-Datensätze pro Zelle (statt pro ganzer
	 * Zeile) gebildet werden sollen.
	 * @return {@code true}, wenn es sich um eine NoChange-Markierung handelt, sonst {@code false}
	 */
	public abstract boolean isNoChange();

	/**
	 * Gibt {@code true} zurück, wenn es sich bei dem übergebenen Wert um eine NoChange-Markierung handelt. NoChange-Markierungen dieser Art
	 * werden PuA-Seitig generiert, wenn es sich um ein Zustandsprotokoll handelt und die NoChange-Datensätze pro Zelle (statt pro ganzer
	 * Zeile) gebildet werden sollen.
	 * @param data Data-objekt, das geprüft werden soll
	 * @return {@code true}, wenn es sich um eine NoChange-Markierung handelt, sonst {@code false}
	 */
	public static boolean isNoChange(Data data){
		if(data instanceof ProtocolData) {
			return ((ProtocolData) data).isNoChange();
		}
		return false;
	}

}
