/*
 * Copyright 2005-2008 by beck et al. projects GmbH, Munich
 * Copyright 2009 by Kappich Systemberatung, Aachen
 *
 * This file is part of de.bsvrz.pua.prot.
 *
 * de.bsvrz.pua.prot is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * de.bsvrz.pua.prot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.pua.prot.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Contact Information:
 * Kappich Systemberatung
 * Martin-Luther-Straße 14
 * 52062 Aachen, Germany
 * phone: +49 241 4090 436
 * mail: <info@kappich.de>
 */

package de.bsvrz.pua.prot.processing;

import de.bsvrz.pua.prot.processing.util.BaseDataSet;
import de.bsvrz.sys.funclib.losb.util.Tuple;

import java.util.HashSet;
import java.util.List;
import java.util.Set;

/**
 * XXX Statusberechung funktioniert nur im Sekundenbereich, Testfälle anpassen! Berechnet den Status der Protokollerstellung.
 *
 * @author beck et al. projects GmbH
 * @author Martin Hilgers
 * @version $Revision$ / $Date$ / ($Author$)
 */
public class Status {

	/** Je ein Datenzeitstempel pro Datenidentifikation */
	private long[] _dataTimeStamps;

	/** Zeitbereiche */
	private long[] _endOfInterval;

	/** Startzeit der Zeitintervalle. */
	private long[] _startOfInterval;

	/** Jedes Zeitintervall wird als Vielfaches des kleinsten Zeitintervalls dargestellt. */
	private double[] _intervalFactor;

	/** Gesamtzahl aller zu übertragenen Zeitbereiche und Datenarten. (Unter Berücksichtigung des intervalFactors) */
	private double _numberOfAllIntervals;

	/** Position der Zeitdauer-Attribute im Spalten-Bereich. */
	private final Set<Integer> _ignore;

	/**
	 * @param periods         Zeitbereiche, für die der Status berechnet werden soll. Die Zeitbereiche dürfen sich nicht überschneiden. Der Anfangszeitpunkt eines
	 *                        Zeitintervalls muss kleiner sein als der Endzeitpunkt. Liste wird nicht verändert.
	 * @param ignore          Position der Zeitdauer-Attribute im Spalten-Bereich. Die Liste ist leer, wenn kein Zeitdauer-Attribut verwendet wird.
	 *
	 * @throws IllegalArgumentException Dauer eines Zeitintervalls ist 0.
	 */
	public Status(final List<Tuple<Long, Long>> periods, final Set<Integer> ignore) throws IllegalArgumentException {
		this._ignore = ignore;
		this._dataTimeStamps = null;
		final long[] duration = new long[periods.size()];
		// Kürzestes Zeitintervall finden & Startzeit & Endezeit ablegen
		_endOfInterval = new long[periods.size()];
		_startOfInterval = new long[periods.size()];
		long minInterval = Long.MAX_VALUE;
		int i = 0;
		for(final Tuple<Long, Long> period : periods) {
			_startOfInterval[i] = period.first;
			_endOfInterval[i] = period.last;
			duration[i] = period.last - period.first + 1;
			if(duration[i] == 0) {
				throw new IllegalArgumentException(
						ProcessingErrorMessage.INVALID_PERIOD_DURATION + period.first + " - " + period.last
				);
			}
			if(duration[i] < minInterval) minInterval = duration[i];
			i++;
		}
		// Faktor berechnen & Nenner
		_numberOfAllIntervals = 0;
		_intervalFactor = new double[duration.length];
		for(i = 0; i < duration.length; i++) {
			_intervalFactor[i] = duration[i] / (double)minInterval;
			_numberOfAllIntervals += _intervalFactor[i] * 1;
		}
	}

	/**
	 * Liefert den Status.
	 *
	 * @param baseDataSets Datensätze mit Zeitstempeln.
	 * @param interval Index des Zeitintervalls, in dem die Datensätze von <code>baseData</code> liegen.
	 *
	 * @return Status zwischen 0 und 100.
	 */
	public byte getStatus(final List<BaseDataSet> baseDataSets, final int interval) {
		if(interval >= _endOfInterval.length) {
			return (byte)100;
		}
		else if(interval < 0) return (byte)0;

		if(_dataTimeStamps == null) _dataTimeStamps = new long[baseDataSets.size() - _ignore.size()];

		int timeStampIndex = 0;

		HashSet<Integer> ignored = new HashSet<>(_dataTimeStamps.length);

		// Zeitstempel sammeln
		for(int i = 0; i < baseDataSets.size(); i++) {
			if(_ignore.contains(i)) {
				continue;
			}
			else {
				final BaseDataSet baseDataSet = baseDataSets.get(i);
				if(baseDataSet != null && baseDataSet.getDataTimeStamp() != BaseDataSet.INVALID_VALUE)
				{
					if(this._dataTimeStamps[timeStampIndex] < baseDataSet.getDataTimeStamp())		//prüfen ob der neue DZS größer als der Alte ist
					{
						if(baseDataSet.getDataTimeStamp() <= _endOfInterval[interval]) {
							this._dataTimeStamps[timeStampIndex] = baseDataSet.getDataTimeStamp();
						}
						else
							/*
							 * Falls der neue DZS über der Intervallgrenze liegt (z.B. für Archivdatensätze),
							 * Ende des Zeitintervalls wählen.
							 */ {
							this._dataTimeStamps[timeStampIndex] = _endOfInterval[interval];
						}

					}
				}
				else{
					// Ungültiger Wert. In Set packen, damit nur die Werte benutzt werden, von denen ein aktuelle Zeitstempel bekannt ist.
					ignored.add(timeStampIndex);
				}
				timeStampIndex++;
			}
		}

		// Falls alle Werte keinen Zeitstempel haben, wird die Ignorierliste gelöscht, damit die vorherigen Ergebnisse weiterbenutzt werden.
		if(ignored.size() == _dataTimeStamps.length) ignored.clear();
		
		double status;
		long tempLong = 0;

		// Schritt 1
		for(int i = 0, dataTimeStampsLength = _dataTimeStamps.length; i < dataTimeStampsLength; i++) {

			if(ignored.contains(i)) continue;
			
			final long _dataTimeStamp = _dataTimeStamps[i];
			/*
			 * Abfrage ist nötig für Onlinedaten: Es können Zeitrück- & -vorsprünge auftreten
			 */
			if(_dataTimeStamp <= _endOfInterval[interval] && _dataTimeStamp >= _startOfInterval[interval]) {
				tempLong += _dataTimeStamp - _startOfInterval[interval];
			}
			else if(_dataTimeStamp >= _startOfInterval[interval]) {
				tempLong += _endOfInterval[interval] - _startOfInterval[interval];
			}
		}
		status = tempLong * _intervalFactor[interval] / (_endOfInterval[interval] - _startOfInterval[interval]) / (_dataTimeStamps.length - ignored.size());

		// Schritt 2
		double tempDouble = 0;
		for(int i = 0; i < interval; i++)
		{
			// diese Intervalle wurden schon vollständig übertragen
			tempDouble += _intervalFactor[i];
		}

		// Schritt 3
		status = (status + tempDouble) / _numberOfAllIntervals * 100;

		if(status < 0) {
			return 0;
		}
		else if(status > 100) {
			return 100;
		}
		else {
			return (byte)status;
		}
	}
}
