/*
 * Copyright 2005-2008 by beck et al. projects GmbH, Munich
 * Copyright 2009 by Kappich Systemberatung, Aachen
 *
 * This file is part of de.bsvrz.pua.prot.
 *
 * de.bsvrz.pua.prot is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * de.bsvrz.pua.prot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.pua.prot.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Contact Information:
 * Kappich Systemberatung
 * Martin-Luther-Straße 14
 * 52062 Aachen, Germany
 * phone: +49 241 4090 436
 * mail: <info@kappich.de>
 */

package de.bsvrz.pua.prot.processing.archivebuffer;

import com.google.common.collect.*;
import de.bsvrz.dav.daf.accessControl.UserInfo;
import de.bsvrz.dav.daf.main.DataDescription;
import de.bsvrz.dav.daf.main.archive.ArchiveData;
import de.bsvrz.dav.daf.main.archive.ArchiveDataKindCombination;
import de.bsvrz.dav.daf.main.archive.ArchiveDataQueryResult;
import de.bsvrz.dav.daf.main.archive.ArchiveDataSpecification;
import de.bsvrz.dav.daf.main.archive.ArchiveDataStream;
import de.bsvrz.dav.daf.main.archive.ArchiveOrder;
import de.bsvrz.dav.daf.main.archive.ArchiveQueryPriority;
import de.bsvrz.dav.daf.main.archive.ArchiveRequestManager;
import de.bsvrz.dav.daf.main.archive.ArchiveRequestOption;
import de.bsvrz.dav.daf.main.archive.ArchiveTimeSpecification;
import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.pua.prot.Protocols;
import de.bsvrz.pua.prot.processing.ProcessingErrorMessage;
import de.bsvrz.sys.funclib.dataIdentificationSettings.DataIdentification;
import de.bsvrz.sys.funclib.debug.Debug;
import de.bsvrz.sys.funclib.losb.exceptions.FailureException;
import de.bsvrz.sys.funclib.losb.exceptions.LoggerException;

import java.io.IOException;
import java.util.*;

/**
 * Kapselt alle Informationen die zum Stellen von Archivanfragen nötig sind. Die Klasse ermöglicht es, Standardparameter für Archivanfragen festzulegen. Es wird
 * nur einmal eine Archivanfrage pro Datenidentifikation gestellt. Dabei wird jedoch vorausgesetzt, dass die {@link
 * de.bsvrz.pua.prot.processing.archivebuffer.RealArchiveElement} Objekte alle direkt hintereinander nach dem gleichen Wert anfragen. Es ist nicht möglich, dass
 * eines dieser Objekte z.B. drei Werte aus dem Archivsystem holt, solange ein anderes Objekt nur einen Wert abgeholt hat. Dieses Fehlverhalten wird von dieser
 * Klasse <b>nicht</b> abgefangen! Durch Aufruf der {@link #setArchiveDataKinds(ArchiveDataKindCombination)} und {@link #setTimeSpec(ArchiveTimeSpecification)} werden die
 * bisherigen Anmeldungen verworfen. Für Relativanfragen gilt diese Regelung nicht: Hier wird immer eine neue Relativanfrage gestellt. Durch den Aufruf von
 * {@link #request(Object,DataDescription,SystemObject)} wird keine Archivanfrage gestellt, sondern es werden alle nötigen Archivanfragen gesammelt. Erst der
 * Aufruf von {@link #execute()} fasst alle bisher gesammelten Anfragen pro Objekttyp in einer Archivanfrage zusammen. (D.h. alle für alle Datenidentifikationen
 * mit dem gleichen Objekttyp wir eine Anfrage gestellt)  Die Methoden dieser Klasse sind threadsafe.
 *
 * @author beck et al. projects GmbH
 * @author Martin Hilgers
 * @version $Revision$ / $Date$ / ($Author$)
 */
public class ArchiveUser {

	/** Debug Ausgaben */
	private static final Debug _debug = Debug.getLogger();

	/** Anfrageschnittstelle zum Archiv */
	private final ArchiveRequestManager _arm;

	/** Zeitbereich der abgefragt wird */
	private ArchiveTimeSpecification _timeSpec;

	/** Datentyp (online, nachgeliefert, etc.) */
	private ArchiveDataKindCombination _dataKinds;

	/** Sortierung der nachgelieferten Daten */
	private final ArchiveOrder _sortOrder;

	/** Anfrageoptionen: Zustand oder Deltaanfrage */
	private final ArchiveRequestOption _requestOptions;

	/** Abbildung Interessenten zu Datenidentifikationen */
	private final HashMap<Object, DataIdentification> _subscribers = new HashMap<>();

	/** Abbildung Datenidentifikationen zu Streams */
	private final HashMap<DataIdentification, ArchiveDataStream> _dids = new HashMap<>();

	/** Zuordnung Streams zu Anzahl Interessenten (je Stream) */
	private final HashMap<DataIdentification, Integer> _numSubscribers = new HashMap<>();

	/** Abbildung Stream zu Anzahl der Werte, die bereits abgeholt wurden */
	private final HashMap<ArchiveDataStream, Integer> _take = new HashMap<>();

	/** Abbildung Stream zu aktuellem Wert */
	private final HashMap<ArchiveDataStream, ArchiveData> _values = new HashMap<>();

	/** Fasst Datenidentifikationen mit dem gleichen ObjektTYP zusammen. DI (mit TYP) -> List<DI(mit Objekt)> */
	private final SetMultimap<DataIdentification, DataIdentification> _diPool = LinkedHashMultimap.create();
	
	/** Wurde die Verbindung unterbrochen? */
	private boolean _aborted = false;
	
	private final UserInfo _userInfo;

	/**
	 * Konstruktor
	 *  @param arm            Anfrageschnittstelle zum Archiv
	 * @param timeSpec       Zeitbereich der abgefragt wird
	 * @param dataKindCombination       Datentyp (online, nachgeliefert, etc.)
	 * @param sortOrder      Sortierung der nachgelieferten Daten
	 * @param requestOptions Anfrageoptionen: Zustand oder Deltaanfrage
	 * @param userInfo       Anfragender Benutzer für Rechteprüfung
	 */
	public ArchiveUser(
			ArchiveRequestManager arm,
			ArchiveTimeSpecification timeSpec,
			ArchiveDataKindCombination dataKindCombination,
			ArchiveOrder sortOrder,
			ArchiveRequestOption requestOptions,
			final UserInfo userInfo) {
		this._arm = arm;
		this._timeSpec = timeSpec;
		this._dataKinds = dataKindCombination;
		this._sortOrder = sortOrder;
		this._requestOptions = requestOptions;
		_userInfo = userInfo;
	}

	/**
	 * Stellt eine Anfrage ans Archivsystem. Falls für die übergebene Datenidentifikation bereits eine Archivanfrage gestellt wurde, wird keine weitere Anfragen
	 * gestellt.
	 *
	 * @param subscriber      Abonnent der Archivdaten
	 * @param dataDescription Datenstruktur
	 * @param object          Objekt
	 */
	public synchronized void request(Object subscriber, DataDescription dataDescription, SystemObject object) {
		//Prüfen ob bereits eine Anfrage besteht:
		DataIdentification di = new DataIdentification(object, dataDescription);

		//Anfrage für neue DI: 1 eintragen
		//Anfrage für die gleiche DI wurde schon gestellt - nur Anzahl Abonnenten erhöhen
		_numSubscribers.merge(di, 1, Integer::sum);

		//Zuordnung Anfrager -> Datenidentifikation ablegen
		_subscribers.put(subscriber, di);
		
		//Anfragen nach Typ zusammenstellen
		DataIdentification typeDi = new DataIdentification(di.getObject().getType(), di.getDataDescription());
		
		_diPool.put(typeDi, di);
	}

	/**
	 * Liefert einen Archivdatensatz. Wichtig: Alle Interessenten einer Datenidentifikation müssen diese Methode hintereinander aufrufen. Beispiel: Interessenten a
	 * und b für Datenidentifikation x. Die Anfragenreihenfolge take(a) take(b) take(a) ist gültig, take(a) take(a) take(b) nicht!
	 *
	 * @param subscriber Interessent
	 *
	 * @return Archivdatensatz.
	 *
	 * @throws InterruptedException  Falls der aufrufende Thread unterbrochen wurde, während auf den nächsten Datensatz gewartet wurde.
	 * @throws IOException           Ein-/Ausgabefehler
	 * @throws IllegalStateException Falls der Ergebnisdatenstrom mit der Methode abort abgebrochen wurde.
	 */
	public synchronized ArchiveData take(Object subscriber) throws InterruptedException, IllegalStateException, IOException {
		ArchiveData ad = null;
		DataIdentification di = _subscribers.get(subscriber);
		if(di != null) {
			ArchiveDataStream ads = _dids.get(di);
			if(ads != null) {
				//Prüfen ob ein neues take ausgeführt werden muss
				int numValuesTaken = _take.get(ads);
				if(numValuesTaken == 0 || numValuesTaken == _numSubscribers.get(di))	//falls take(ads) == 0 -> es wurden noch keine Daten angefordert!
				{
					//Alle Interessenten haben die Daten abgeholt -> neues take ausführen
					ad = ads.take();
					_values.put(ads, ad);
					_take.put(ads, 1);
				}
				else {
					//Es haben noch nicht alle Interessenten die Daten abgeholt
					_take.put(ads, (numValuesTaken + 1));	//Anzahl abgeholter Daten erhöhen
					ad = _values.get(ads);		//Wert holen.
				}
			}
		}
		return ad;
	}

	/**
	 * Liefert genau einen Archivdatensatz mit kleineren-gleichen Datenzeitstempel als dem übergebenen Intervallende
	 *
	 * @param dataDescription Datenstruktur
	 * @param object          Objekt
	 * @param timeStamp       Intervallende
	 *
	 * @return Informationen über die Anfrage
	 */
	public ArchiveDataQueryResult requestRelative(DataDescription dataDescription, SystemObject object, long timeStamp) throws FailureException {

		Protocols.checkPermissions(_userInfo, object, dataDescription);

		ArchiveTimeSpecification ats = new ArchiveTimeSpecification(_timeSpec.getTimingType(), true, 1, timeStamp);
		ArchiveDataSpecification ads = new ArchiveDataSpecification(ats, _dataKinds, _sortOrder, _requestOptions, dataDescription, object);
		enablePidQuery(ads);
		return _arm.request(ArchiveQueryPriority.MEDIUM, ads);
	}

	/**
	 * Setzt die Zeitspezifikation. Beendet alle bisherigen Verbindungen zum Archivsystem.
	 *
	 * @param timeSpec Zeitspezifikation
	 */
	public void setTimeSpec(ArchiveTimeSpecification timeSpec) {
		reset();
		this._timeSpec = timeSpec;
	}

	public ArchiveTimeSpecification getTimeSpec() {
		return _timeSpec;
	}

	/**
	 * Setzt die Datenart, die abgefragt werden soll. Beendet alle bisherigen Verbindungen zum Archivsystem.
	 *
	 * @param dataKindCombination DataKindCombination
	 */
	public void setArchiveDataKinds(final ArchiveDataKindCombination dataKindCombination) {
		reset();
		this._dataKinds = dataKindCombination;
	}

	/** Setzt die Verbindungen zum Archivsystem zurück. Offene Streams werden geschlossen. */
	private synchronized void reset() {
		for(ArchiveDataStream ads : _take.keySet()) {
			ads.abort();
		}

		_numSubscribers.clear();
		_subscribers.clear();
		_take.clear();
		_values.clear();
		_dids.clear();
		_diPool.clear();
		_aborted = true;
	}

	/** Beendet alle ausstehenden Archivanfrage. Es werden sofort alle Archivanfragen beendet, auch diejenigen, die von anderen Objekten gestellt wurden. */
	public void abort() {
		reset();
	}

	/**
	 * Sendet die bisher in {@link #request(Object,DataDescription,SystemObject)} gesammelten Archivanfragen in einer Anfrage an das Archivsystem.
	 *
	 * @throws FailureException     Archivsystem ist nicht erreichbar.
	 * @throws InterruptedException Thread wird abgebrochen
	 */
	public void execute() throws FailureException, InterruptedException {

		Set<DataIdentification> didSet;
		
		synchronized(this) {
			_aborted = false;
			didSet = ImmutableSet.copyOf(_diPool.values());
		}

		for(DataIdentification it : didSet) {
			Protocols.checkPermissions(_userInfo, it.getObject(), it.getDataDescription());
		}

		ArrayList<ArchiveDataSpecification> adSpec = new ArrayList<>(didSet.size());
		for(DataIdentification did : didSet) {
			ArchiveDataSpecification ads = new ArchiveDataSpecification(
					_timeSpec, _dataKinds, _sortOrder, _requestOptions, did.getDataDescription(), did.getObject()
			);
			enablePidQuery(ads);
			adSpec.add(ads);
		}

		//Archivanfrage stellen
		try {
			_debug.finest("Stelle Archivanfrage für:", didSet);
			ArchiveDataQueryResult adq = _arm.request(ArchiveQueryPriority.MEDIUM, adSpec);
			_debug.info("Archivanfrage gestellt, prüfe ob Anfrage erfolgreich.");
			if(!adq.isRequestSuccessful()) {
				_debug.info("Archivanfrage fehlerhaft.");
				throw new FailureException(ProcessingErrorMessage.ARCHIVE_REQUEST_FAILED + adq.getErrorMessage(), LoggerException.WARNING);
			}
			else {
				_debug.info("Archivanfrage erfolgreich");
				if(adq.getStreams().length != adSpec.size()) {
					throw new FailureException(ProcessingErrorMessage.TOO_MANY_STREAMS + adq.getStreams().length, LoggerException.WARNING);
				}
				else {
					synchronized(this) {
						//Streams aufteilen:
						for(int i = 0; i < adq.getStreams().length; i++) {
							ArchiveDataStream ads = adq.getStreams()[i];
							_take.put(ads, 0);    // es wurden noch keine Daten abgeholt
							ArchiveDataSpecification dataSpec = ads.getDataSpecification();
							DataIdentification dataIdentification = new DataIdentification(dataSpec.getObject(), dataSpec.getDataDescription());
							_dids.put(dataIdentification, ads);
						}
						if(_aborted){
							reset();
						}
					}
				}
			}
		}
		catch(IllegalStateException e) {
			throw new FailureException(
					"ArchiveUser.request " + ProcessingErrorMessage.ARCHIVE_COMMUNICATION + e.getMessage(), e, LoggerException.WARNING
			);
		}
	}

	private static void enablePidQuery(final ArchiveDataSpecification ads) {
		if(Protocols._fetchOldObjects){
			try {
				ads.setQueryWithPid();
			}
			catch(NoSuchMethodError ignored){
				// Kompatibilität bei alten DAF
				_debug.warning("Archivanfrage kann historische Objekte nicht berücksichtigen, bitte DAF-Bibliothek aktualisieren.");
			}
		}
	}

}
