/*
 * Copyright 2005-2008 by beck et al. projects GmbH, Munich
 * Copyright 2009 by Kappich Systemberatung, Aachen
 *
 * This file is part of de.bsvrz.pua.prot.
 *
 * de.bsvrz.pua.prot is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * de.bsvrz.pua.prot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.pua.prot.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Contact Information:
 * Kappich Systemberatung
 * Martin-Luther-Straße 14
 * 52062 Aachen, Germany
 * phone: +49 241 4090 436
 * mail: <info@kappich.de>
 */

package de.bsvrz.pua.prot.processing.onlinebuffer;

import de.bsvrz.dav.daf.accessControl.UserInfo;
import de.bsvrz.dav.daf.main.ClientDavInterface;
import de.bsvrz.dav.daf.main.ReceiveOptions;
import de.bsvrz.pua.prot.processing.ProcessingBuffer;
import de.bsvrz.pua.prot.processing.ProcessingInterface;
import de.bsvrz.pua.prot.processing.ValueProvider;
import de.bsvrz.pua.prot.processing.util.BaseDataSet;
import de.bsvrz.pua.prot.util.ProcessingInformation;
import de.bsvrz.sys.funclib.losb.datk.ProtocolType;
import de.bsvrz.pua.prot.util.attributes.AttributeGroupDescription;
import de.bsvrz.pua.prot.util.attributes.RealAttributeDescription;
import de.bsvrz.pua.prot.util.attributes.RealElement;
import de.bsvrz.sys.funclib.losb.exceptions.FailureException;
import de.bsvrz.sys.funclib.losb.util.Tuple;

import java.util.ArrayList;
import java.util.List;

/**
 * Holt benötigte Daten vom Datenverteiler. Meldet sich als einfacher Empfänger beim Datenverteiler für die benötigten Datensätze an. Nach dem Erzeugen des
 * Objektes sollte sofort {@link #init()} aufgerufen werden. Dadurch werden Objekte erzeugt, die sich beim DAV als Empfänger für Online-Daten anmelden.
 *
 * @author beck et al. projects GmbH
 * @author Martin Hilgers
 * @version $Revision$ / $Date$ / ($Author$)
 */
public class OnlineBuffer extends ProcessingBuffer {

	/** Zeitpunkt, an dem sich der Online ProcessingBuffer auf alle Datenidentifikationen angemeldet hat. */
	private long _doneSubscribing;

	/** Zeigt an ob die Datensammlung abgeschlossen ist. */
	private boolean _doneCollecting = false;

	/** Intervalle, für die {@link ProcessingBuffer#nextInterval(long)} bereits aufgerufen wurde. */
	private int _countedIntervals = 0;

	/** Objekt das den Eingang von neuen Daten überwacht. */
	private NewDataCollector _newDataColllector;

	/** Objekt das die Empfänger verwaltet. */
	private OnlineDataReceiverSet _receiverSet;

	/** Zeigt an ob der Zeitpunkt des letzten Intervalls überschritten wurde. */
	private boolean _lastIntervalReached = false;

	private long _lastTimeStamp;
	
	private final UserInfo userInfo;

	/**
	 * {@link #init()} sollte unverzüglich nach Erzeugen des Objekts aufgerufen werden.
	 *
	 * @param dav       Verbindung zum Datenverteiler.
	 * @param processor Objekt, das bei Ankunft von Daten benachrichtigt wird.
	 * @param pi        Informationen zur Datenaufbereitung.
	 * @param periods   Zeitbereiche, in denen Daten gesammelt werden sollen. Die Zeitbereiche müssen bereits sortiert und zusammengefasst sein!
	 * @param startPeriodIndex StartPeriodIndex
	 * @param scriptName Bezeichnung des Skripts zur Verwendung in Fehlerausgaben.
	 * @param userInfo
	 * @throws FailureException Fehler bei der Kommunikation mit der Konfiguration.
	 */
	public OnlineBuffer(final ClientDavInterface dav, final ProcessingInterface processor, final ProcessingInformation pi, final List<Tuple<Long, Long>> periods, final int startPeriodIndex, final String scriptName, final UserInfo userInfo)
			throws FailureException {
		super(dav, processor, pi, periods, startPeriodIndex, scriptName);
		this.userInfo = userInfo;
		this.setAggregate(false);
		_newDataColllector = new NewDataCollector();
	}

	/**
	 * Erzeugt zu jedem Element ein Objekt zum Typ {@link RealOnlineElement} und startet den Thread. Diese melden sich beim Datenverteiler als Empfänger für
	 * Online-Daten an. Falls es zu einem Fehler beim anmelden als Empfänger beim Dav kommt, wird der Thread nicht gestartet, und {@link ProcessingBuffer#abort()}
	 * aufgerufen.
	 *
	 * @see de.bsvrz.pua.prot.processing.ProcessingBuffer#init()
	 */
	@Override
	public void init() throws FailureException {
		try {
			if(!_periods.isEmpty()) {
				boolean withDelayedData = false;
				if(_processingInformation.getArchiveDataKinds().isOnlineDelayed()){
					withDelayedData = true;
				}

				ProtocolType protocolType = _processingInformation.getProtocolType();

				final ReceiveOptions receiveOptions = new ReceiveOptions(protocolType == ProtocolType.DeltaProtocol, withDelayedData);
				// Reale Elemente initialisieren.
				_receiverSet = new OnlineDataReceiverSet(_connection, receiveOptions, _periods, _newDataColllector, userInfo);
				for(final RealElement realElement : _processingInformation.getAllRealElements()) {
					if(realElement instanceof RealAttributeDescription) {
						final RealAttributeDescription re = (RealAttributeDescription)realElement;
						final RealOnlineAttribute roa = new RealOnlineAttribute(_receiverSet, re, protocolType, _processingInformation.isShown(realElement));
						_realElements.put(re, roa);
					}
					else if(realElement instanceof AttributeGroupDescription) {
						final AttributeGroupDescription atg = (AttributeGroupDescription)realElement;
						final RealOnlineAttributeGroup roa = new RealOnlineAttributeGroup(_receiverSet, atg, protocolType, _processingInformation.isShown(realElement));
						_realElements.put(atg, roa);
					}
				}
				// Temporäre Elemente initialisieren.
				setTempAttributes();
				this.start();
			}
			else {
				setDone();
			}
			_doneSubscribing = System.currentTimeMillis();
		}
		catch(FailureException e) {
			this.abort();
			throw e;
		}
	}

	/**
	 * Gibt dem Zeitpunkt zurück, an dem der Online ProcessingBuffer mit dem Sammeln der Daten begonnen hat.
	 *
	 * @return Zeitpunkt ab dem Daten gesammelt werden.
	 */
	public long getStartTime() {
		return _doneSubscribing;
	}

	/**
	 * Online Daten werden nicht aggregiert. Daher liefert diese Methode <b>immer</b> <code>true</code> zurück.
	 *
	 * @return Immer <code>true</code>
	 *
	 * @see de.bsvrz.pua.prot.processing.ProcessingBuffer#applyAggregations(List)
	 */
	@Override
	public boolean applyAggregations(
			final List<BaseDataSet> baseData) {
		return true;
	}

	/** @see de.bsvrz.pua.prot.processing.ProcessingBuffer#hasData() */
	@Override
	public boolean hasData() {
		return (_buffer.size() >= 2 || (_doneCollecting && !_buffer.isEmpty()));
	}

	/**
	 * Überprüft, ob das Ende der zu überprüfenden Zeitbereiche erreicht ist und alle Daten übertragen wurden.
	 *
	 * @see de.bsvrz.pua.prot.processing.ProcessingBuffer#isDoneCollecting()
	 */
	@Override
	protected boolean isDoneCollecting() {
		if(_doneCollecting || isAbort()) {
			return true;
		}
		else {
			/*
			 * Erst Datenaufbereitung abschließen wenn sowohl das letzte Zeitintervall abgeschlossen wurde, als auch
			 * keine weiteren Daten mehr in der 'Pipeline' vorliegen.
			 */
			_doneCollecting = checkPeriod() && !_newDataColllector.hasMore();
			if(_doneCollecting) {
				/*
				 * Letztes Intervall wurde überschritten -> sicherstellen, dass alle Daten verarbeitet wurden. Falls
				 * nicht: Empfang von Daten melden.
				 */
				_doneCollecting = !_receiverSet.dataAvailable();
				if(!_doneCollecting) _newDataColllector.newDataAvailable();
			}
			return _doneCollecting;
		}
	}

	/** @see de.bsvrz.pua.prot.processing.ProcessingBuffer#isDone() */
	@Override
	public boolean isDone() {
		return _done || (_doneCollecting && _buffer.isEmpty() && _dataSetBuilder.isDone());
	}

	/**
	 * Überprüft ob bereits ein neues Intervall angebrochen wurde.
	 *
	 * @return <code>true</code> falls das letzte Intervall zeitlich bereits überschritten.
	 */
	private boolean checkPeriod() {
		if(_lastIntervalReached) return true;
		/*
		 * überprüfen, in welchem Zeitinterval man sich befindet, und ggf. nextStatusInterval aufrufen. (die Schleife
		 * ist notwendig, da sich das Intervall mehrfach geändert haben kann seitdem checkPeriods das letzte Mal
		 * aufgerufen wurde.
		 */
		final long now = System.currentTimeMillis();
		for(; _countedIntervals < _periods.size() && _periods.get(_countedIntervals).last < now; _countedIntervals++) {
			nextInterval(_periods.get(_countedIntervals).last);
		}

		if(_countedIntervals == _periods.size()) {
			_lastIntervalReached = true;
			return true;
		}
		else {
			return false;
		}
	}

	/**
	 * Liefert immer false zurück.
	 *
	 * @see de.bsvrz.pua.prot.processing.ProcessingBuffer#isListAggregation()
	 */
	@Override
	protected boolean isListAggregation() {
		return false;
	}

	/**
	 * Online gesammelte Daten werden nicht aggregiert. Daher tut diese Methode nichts.
	 *
	 * @see de.bsvrz.pua.prot.processing.ProcessingBuffer#storeAggregatedData(byte)
	 */
	@Override
	public void storeAggregatedData(
			final byte status) {
	}

	/**
	 * Berechnet die Zeit bis zum Ende des nächsten Intervalls.
	 * @return Zeit
	 */
	protected long getMillisToEndOfInterval() {
		long waitTime = -1;
		for(int i = 0; i < _periods.size() && waitTime <= 0; i++) {
			waitTime = _periods.get(i).last - System.currentTimeMillis();
		}

		if(waitTime <= 0) waitTime = 1;
		return waitTime;
	}

	/**
	 * XXX MH: Merge sollte schon im Konstruktor von imdsBuffer festgelegt werden.
	 *
	 * @see de.bsvrz.pua.prot.processing.ProcessingBuffer#processNewData(ValueProvider[], ArrayList)
	 */
	@Override
	protected void processNewData(final ValueProvider[] elements, final ArrayList<ValueProvider> winners) throws FailureException, InterruptedException {
		if(waitForNewData()) {
			// Werte holen
			if(winners.isEmpty()) winners.addAll(_realElements.values());
			for(final ValueProvider provider : winners) {
				/* Datensatz holen. */
				if(provider.take(1)) provider.updateDataInformation(); // Markierungsinformation updaten, falls Daten vorliegen.
			}

			/*
			 * >0 falls Gewinner im Spaltenbereich, <0 falls Gewinner woanders, 0 falls kein Gewinner gefunden
			 */
			final int hasWinners;
			// Element mit minimalem DZS auswählen
			hasWinners = getWinners(elements, winners);
			if(hasWinners > 0) // Im Spaltenbereich hat sich was getan -> Daten aufbereiten
			{
				// Minimalen Zeitstempel merken den nach oben ziehen...
				final long winnerTimeStamp = winners.get(0).getMostRecentTimeStamp();

				if(winnerTimeStamp != _lastTimeStamp){
					// Wenn sich der Zeitstempel geändert hat bisherige Daten senden. (Zusammenfassen von zusammengehörigen Daten)
					_dataSetBuilder.createIntermediateDataSet(false);
					_lastTimeStamp = winnerTimeStamp;
				}
				else{
					// Ansonsten alten Datensatz verwerfen: Zu der aktuellen Zeit gibt es neue Daten, also wird der bisherige Datensatz nicht mehr gebraucht.
					// Ein neuer mit aktuelleren Daten wird sofort nach diesem else-Block angelegt.
					_dataSetBuilder.throwAway();
				}

				// Erst danach neuen Datensatz erstellen
				_dataSetBuilder.collectData(getInsertEmpty(), winnerTimeStamp);
			}
			winners.clear(); // immer alle Elemente nach neuen Daten abfragen
		}
	}

	/**
	 * Wartet, bis neue Daten eingehen, oder das Ende eines Protokoll-Intervalls erreicht wurde.
	 *
	 * @return <code>true</code> falls neue Daten eingegangen sind.
	 */
	private boolean waitForNewData() throws InterruptedException {
		synchronized(_newDataColllector) {
			if(_newDataColllector.hasMoreData()) {
				return true;
			}
			else {
				_newDataColllector.wait(getMillisToEndOfInterval());
				return false;
				/*
				 * Falls in der Zwischenzeit Daten eingetroffen sind, werden diese in der nächsten Runde bearbeitet.
				 */
			}
		}
	}
}
