/*
 * Copyright 2005-2008 by beck et al. projects GmbH, Munich
 * Copyright 2009 by Kappich Systemberatung, Aachen
 *
 * This file is part of de.bsvrz.pua.prot.
 *
 * de.bsvrz.pua.prot is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * de.bsvrz.pua.prot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.pua.prot.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Contact Information:
 * Kappich Systemberatung
 * Martin-Luther-Straße 14
 * 52062 Aachen, Germany
 * phone: +49 241 4090 436
 * mail: <info@kappich.de>
 */

package de.bsvrz.pua.prot.interpreter.semantics;

import antlr.collections.AST;
import de.bsvrz.dav.daf.main.config.ConfigurationException;
import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.dav.daf.main.config.SystemObjectType;
import de.bsvrz.pua.prot.exceptions.CriticalParserException;
import de.bsvrz.pua.prot.exceptions.ScopeException;
import de.bsvrz.pua.prot.exceptions.SemanticErrorException;
import de.bsvrz.pua.prot.interpreter.InterpreterErrorMessage;
import de.bsvrz.pua.prot.util.MemberCheck;
import de.bsvrz.pua.prot.util.ProcessingParameter;
import de.bsvrz.pua.prot.util.Qualifier;
import de.bsvrz.pua.prot.util.attributes.AttributeGroupDescription;
import de.bsvrz.pua.prot.util.attributes.RealAttributeDescription;
import de.bsvrz.sys.funclib.losb.messages.ErrorMessage;

import java.util.*;

/**
 * Überprüft die Semantik von Attributgruppen.
 *
 * @author beck et al. projects GmbH
 * @author Martin Hilgers
 * @version $Revision$ / $Date$ / ($Author$)
 */
@SuppressWarnings("deprecation")
public class AttributeGroupSemantics implements SemanticChecks {

	private Semantics _semantics;

	private final RealAttributeSemantics _realAttributeSemantics;

	private ScopeSemantics _scopeSemantics;

	private MemberCheck _check;

	private AttributeSemantics _attributeSemantics;

	/**
	 * @param semantics Semantics
	 * @param scopeSemantics ScopeSemantics
	 * @param attributeSemantics AttributeSemantics
	 * @param realAttributeSemantics RealAttributeSemantics
	 */
	public AttributeGroupSemantics(
			Semantics semantics, ScopeSemantics scopeSemantics, AttributeSemantics attributeSemantics, final RealAttributeSemantics realAttributeSemantics) {
		this._semantics = semantics;
		_realAttributeSemantics = realAttributeSemantics;
		this._check = semantics.getCheck();
		this._scopeSemantics = scopeSemantics;
		this._attributeSemantics = attributeSemantics;
	}

	/**
	 * Füg eine Attributgruppe in die zu verarbeitenden Daten ein
	 *
	 * @param ast        Knoten (wird nur für Zeilenangabe in der Fehlermeldung benötigt)
	 * @param parts      Qualifizierung der Attributgruppe
	 * @param columnName Name der Spalte
	 * @param padData    <code>true</code> falls Daten durch die Datenaufbereitung aufgefüllt werden sollen.
	 *
	 * @throws CriticalParserException Falls es zu einem Kommunikationsfehler kommt oder die Attributgruppe nicht korrekt qualifiziert wurde.
	 */
	public void addAttributeGroup(AST ast, List<String> parts, String columnName, boolean padData) throws CriticalParserException {
		try {
			if(parts == null || parts.isEmpty()) throw new ScopeException(InterpreterErrorMessage.NO_SCOPE);

			AttributeGroupDescription atgDesc[];
			atgDesc = createAttributeGroupDescription(parts, padData);

			for(final AttributeGroupDescription anAtgDesc : atgDesc) {
				anAtgDesc.setColumnName(columnName);
				_semantics.pi.addElement(anAtgDesc);
			}
		}
		catch(CriticalParserException criticalException) {
			criticalException.line = ast.getLine();
			throw criticalException;
		}
	}

	/**
	 * Erzeugt Attributgruppenbeschreibungen. Dabei wird die Qualifizierung der Attributgruppe überprüft. Wird ein Alias verwendet, so wird er aufgelöst. Daher ist
	 * es moeglich, dass mehrere Attributgruppen zurueckgegeben werden: Wenn einem Alias mittels Pseudoobjekt-Definition mehrere Objekte zugewiesen wurden. (=>
	 * Wird ein Alias verwendet, werden alle durch die Pseudoobjektdefinitionen erzeugten Attributgruppe zurückgegeben.) Erhöht zudem den Zähler für reale
	 * Attribute, falls die Attributgruppe Attribute enthält.
	 *
	 * @param parts   Die Qualifizierung der Attributgruppe. Die Liste wird durch diese Methode nicht geändert.
	 * @param padData <code>true</code> falls Daten durch die Datenaufbereitung aufgefüllt werden sollen.
	 *
	 * @return Array mit den gebildeten Attributgruppenbeschreibungen
	 *
	 * @throws CriticalParserException Falls es zu einem Fehler kam (Kommuniktation, fehlerhafte Qualifizierung)
	 */
	private AttributeGroupDescription[] createAttributeGroupDescription(List<String> parts, boolean padData) throws CriticalParserException {
		AttributeGroupDescription atgDesc = new AttributeGroupDescription(_semantics.pi.getSimVar(), padData);

		/**
		 * Folgende Qualifizierungen sind möglich
		 * parts.size():	Pfad
		 * 1:	atg:asp				(benutze Qualifizierung aus Stack)
		 * 2:	asp					(benutze Qualifizierung aus Stack)
		 * 3:	Obj:Atg:Asp
		 *
		 * 4:	Obj					(benutze Qualifizierung aus Stack)
		 * 5:	Obj:Atg				(benutze Qualifizierung aus Stack)
		 * 6:	Obj:Asp				(benutze Qualifizierung aus Stack)
		 */

		Qualifier quali = null;

		//überprüfen ob ein Alias verwendet wird
		final String node = parts.get(0);
		final String aliasType = _semantics.symbolTable.getAlias(node);

		final String[] objects;

		if(aliasType != null) {
			//Es wird ein Alias verwendet
			//Alias durch Objekttyp austauschen
			parts.set(0, aliasType);

			// Die zugehörigen Pseudo-Objekte holen
			objects = _semantics.symbolTable.getAliasObjects(node);
		}
		else {
			objects = null;
		}

		//prüfen was für einen Typ das erste parts element hat
		int type = _check.whatIs(parts.get(0));
		//Objekt oder Objekttyp

		if(type == MemberCheck.OBJECT || type == MemberCheck.OBJECTTYPE) {
			if(parts.size() == 3)	//atg ist vollständig qualifiziert
			{
				atgDesc.qualifier.object = _check.getObject(parts.get(0));
				atgDesc.qualifier.attributeGroup = _check.getAttributeGroup(parts.get(1));
				atgDesc.qualifier.aspect = _check.getAspect(parts.get(2));
				if(atgDesc.qualifier.aspect == null) atgDesc.qualifier.freeAspect = parts.get(2);
			}
			else {
				//atg ist nicht vollständig qualifiziert -> Obj & ggf. Asp aus Stack-Qualifizierung holen
				quali = _scopeSemantics._objectToScope.get(parts.get(0));
				if(quali != null) {
					atgDesc.qualifier = quali.clone();
					if(parts.size() == 2) {
						String second = parts.get(1);
						type = _check.whatIs(second);
						if(type == MemberCheck.ATTRIBUTEGROUP) {
							atgDesc.qualifier.attributeGroup = _check.getAttributeGroup(second);
						}
						else if(type == MemberCheck.ASPECT) {
							atgDesc.qualifier.aspect = _check.getAspect(second);
							atgDesc.qualifier.freeAspect = null;
						}
						else if(type == MemberCheck.FREE_ASPECT) {
							atgDesc.qualifier.freeAspect = second;
							atgDesc.qualifier.aspect = null;
						}
					}
					else if(parts.size()
					        != 1)		//wenn size == 1 -> dann ist nur das objekt angegeben -> braucht nix mehr getan zu werden, quali entspricht Qualifizierung aus dem mit-Block
					{
						throw new SemanticErrorException(InterpreterErrorMessage.INVALID_SCOPE);
					}
				}
				else {
					throw new SemanticErrorException(InterpreterErrorMessage.INVALID_SCOPE);
				}
			}
		}
		else {
			quali = _scopeSemantics.peek();
			if(quali == null) throw new ScopeException(InterpreterErrorMessage.INVALID_SCOPE + parts.get(0));

			atgDesc.qualifier = quali.clone();
			//Aspekt
			if(type == MemberCheck.ASPECT && parts.size() == 1) {
				atgDesc.qualifier.aspect = _check.getAspect(parts.get(0));
				atgDesc.qualifier.freeAspect = null;
			}
//Freier Aspekt
			else if(type == MemberCheck.FREE_ASPECT && parts.size() == 1) {
				atgDesc.qualifier.freeAspect = parts.get(0);
				atgDesc.qualifier.aspect = null;
			}
//Attributgruppe ohne Aspekt
			else if(type == MemberCheck.ATTRIBUTEGROUP && parts.size() == 1) {
				atgDesc.qualifier.attributeGroup = _check.getAttributeGroup(parts.get(0));
			}
//Attributgruppe mit Aspekt
			else if(type == MemberCheck.ATTRIBUTEGROUP && parts.size() == 2) {
				atgDesc.qualifier.attributeGroup = _check.getAttributeGroup(parts.get(0));
				type = _check.whatIs(parts.get(1));
				if(type == MemberCheck.ASPECT) {
					atgDesc.qualifier.aspect = _check.getAspect(parts.get(1));
					atgDesc.qualifier.freeAspect = null;
				}
				else if(type == MemberCheck.FREE_ASPECT) {
					atgDesc.qualifier.freeAspect = parts.get(1);
					atgDesc.qualifier.aspect = null;
				}
				else {
					throw new ScopeException(InterpreterErrorMessage.ASPECT_EXPECTED + parts.get(1));
				}
			}
			else {
				throw new ScopeException(InterpreterErrorMessage.OBJECT_OR_OBJECTTYPE_EXPECTED + parts.get(0));
			}
		}

		// überprüfen, ob das eine gültige Qualifizierung ergibt
		atgDesc.qualifier.valildate();
		
		_realAttributeSemantics.checkAdd(aliasType != null, atgDesc.qualifier.object);

		//überprüfen, ob die Attributgruppe Attribute enthält. Wenn ja, Zähler erhöhen, sonst Fehler
		try {
			if(atgDesc.qualifier.attributeGroup.getAttributes().size() == 0) {
				throw new SemanticErrorException(InterpreterErrorMessage.ATG_WO_ATTRIBUTES + atgDesc.qualifier.attributeGroup.getPid());
			}
			_attributeSemantics.incAttributeCount();
		}
		catch(ConfigurationException e) {
			throw new CriticalParserException(ErrorMessage.COMMUNICATION);
		}

		AttributeGroupDescription[] result;
		//falls ein Alias verwendet wurde, jetzt die jeweiligen Attribute bilden. Da dem Alias in der Pseudoobjekt-Definition nur Objekte vom gerade überprüften Typ zugeordnet werden können, stimmt deren Qualifizierung automatisch 
		if(objects != null) {
			result = new AttributeGroupDescription[objects.length];
			for(int i = 0; i < objects.length; i++) {
				AttributeGroupDescription temp = atgDesc.clone();

				//Alias-Objekt gegen konkretes Objekt austauschen
				final String objectPid = objects[i];
				if(objectPid != null && !objectPid.isEmpty()) {
					SystemObject object = _check.getObject(objectPid);
					if(object == null) throw new SemanticErrorException(InterpreterErrorMessage.INVALID_OBJECT_PID + objects[i]);
					if(!object.isOfType(aliasType)) {
						throw new SemanticErrorException("Der Alias-Typ \"" + aliasType + "\" des Aliases \"" + node + "\" konnte nicht durch das Objekt \"" + object + "\" vom Typ \"" + object
								.getType() + "\" ersetzt werden.");
					}
					temp.qualifier.object = object;
					temp.setIndex(i);
					result[i] = temp;
				}
				else{
					temp.qualifier.object = null;
					temp.setIndex(i);
					result[i] = temp;
				}
			}
		}
		else {	//kein Alias -> nur eine Attributbeschreibung
			result = new AttributeGroupDescription[1];
			result[0] = atgDesc;
			//	Objekt gegen konkretes Objekt austauschen
			SystemObject temp = _check.getObject(result[0].qualifier.object);

			// falls keine Bindung existiert, dann Fehler auswerfen, falls das Protokoll ausgewertet werden soll!
			if(temp instanceof SystemObjectType && _semantics.processScript) {
				SystemObject mainObject = _semantics.getSymbolTable().getMainObject();
				if(mainObject == null){
					throw new SemanticErrorException("Es wurde kein Hauptobjekt übergeben, um dem Typ \"" + temp + "\" aufzulösen.");
				}
				throw new SemanticErrorException("Der Typ \"" + temp + "\" konnte nicht durch das Hauptobjekt \"" + mainObject + "\" vom Typ \"" + mainObject.getType() + "\" ersetzt werden.");			}

			result[0].qualifier.object = temp;
		}
		return result;
	}

	/**
	 * Tut nichts.
	 *
	 * @see de.bsvrz.pua.prot.interpreter.semantics.SemanticChecks#cleanUp(de.bsvrz.pua.prot.util.ProcessingParameter,boolean)
	 */
	public void cleanUp(ProcessingParameter pp, boolean processScript) throws SemanticErrorException {
	}
}
