/*
 * Copyright 2005-2008 by beck et al. projects GmbH, Munich
 * Copyright 2009 by Kappich Systemberatung, Aachen
 *
 * This file is part of de.bsvrz.pua.prot.
 *
 * de.bsvrz.pua.prot is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * de.bsvrz.pua.prot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.pua.prot.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Contact Information:
 * Kappich Systemberatung
 * Martin-Luther-Straße 14
 * 52062 Aachen, Germany
 * phone: +49 241 4090 436
 * mail: <info@kappich.de>
 */

package de.bsvrz.pua.prot.interpreter.semantics;

import antlr.collections.AST;
import de.bsvrz.dav.daf.main.config.Aspect;
import de.bsvrz.dav.daf.main.config.AttributeGroup;
import de.bsvrz.dav.daf.main.config.ConfigurationException;
import de.bsvrz.dav.daf.main.config.DataModel;
import de.bsvrz.pua.prot.exceptions.CriticalParserException;
import de.bsvrz.pua.prot.exceptions.ScopeException;
import de.bsvrz.pua.prot.exceptions.SemanticErrorException;
import de.bsvrz.pua.prot.interpreter.InterpreterErrorMessage;
import de.bsvrz.pua.prot.util.MemberCheck;
import de.bsvrz.pua.prot.util.ProcessingParameter;
import de.bsvrz.pua.prot.util.Qualifier;
import de.bsvrz.sys.funclib.losb.messages.ErrorMessage;

import java.util.*;

/**
 * Semantische Überprüfung und Hilfsklassen für das Mit-Konstrukt.
 *
 * @author beck et al. projects GmbH
 * @author Martin Hilgers
 * @version $Revision$ / $Date$ / ($Author$)
 */
@SuppressWarnings("deprecation")
public class ScopeSemantics implements SemanticChecks {

	private MemberCheck _check;

	private DataModel _model;

	/** Stack, oberstes Element enthält aktuell gültige Qualifizierung */
	private Stack<Qualifier> _pathStack;

	/**
	 * Zum schnellen auffinden einer Qualifizierung.<br> [x, y] x: Objekt-pid, y, Qualifizierung.<br> Wird verwendet, da im Mit-Konstrukt durch die Angabe eines
	 * Objekttyps auf eine verschattete Qualifizierung zurückgegriffen werden kann.
	 */
	Hashtable<String, Qualifier> _objectToScope;

	/** Simulationsvariante */
	private short _simVar;

	/** @param semantics semantics  */
	public ScopeSemantics(Semantics semantics) {
		_check = semantics.getCheck();
		_model = semantics.model;
		_simVar = semantics.pi.getSimVar();
		_objectToScope = new Hashtable<>();
		_pathStack = new Stack<>();
	}

	/**
	 * Holt das oberste Stackelement ohne es vom Stack zu löschen.
	 *
	 * @return Oberstes Element des Stacks oder null falls keins vorhanden
	 */
	public Qualifier peek() {
		if(!_pathStack.isEmpty()) {
			return _pathStack.peek();
		}
		else {
			return null;
		}
	}

	/**
	 * Holt das oberste Stackelement und entfernt es vom Stack.
	 *
	 * @return Oberstes Element des Stacks oder null falls keins vorhanden ist
	 */
	public Qualifier pop() {
		if(!_pathStack.isEmpty()) {
			_objectToScope.remove(_pathStack.peek().object.getPid());
			return _pathStack.pop();
		}
		else {
			return null;
		}
	}

	/**
	 * Überprüft Qulifizierung und legt sie auf den Stack. Es werden nur gültige Qualifizierungen abgelegt. Eine gültige Qualifizierung beginnt immer mit einem
	 * Objekt oder Objekttyp:<br> <code>Qualifizierung := Objekt | Objekttyp [Attributgruppe [ Aspekt ]]</code><br> Überzählige Elemente werden ignoriert.
	 *
	 * @param path  Knoten (nur für Zeilenangaben in den Fehlermeldungen)
	 * @param parts Qualifizierung
	 *
	 * @throws CriticalParserException Falls es zu einem Kommunikationsfehler kommt oder falls die Qualifizierung nicht gültig ist
	 */
	public void push(AST path, List<String> parts) throws CriticalParserException {
		try {
			if(parts == null || parts.isEmpty()) throw new ScopeException(InterpreterErrorMessage.NO_SCOPE + parts);

			Qualifier quali = new Qualifier(_simVar);

			/**
			 * Folgende Qualifizierungen sind möglich
			 * parts.size():	Pfad
			 * 1:				Objekt
			 * 2:				Objekt:Attributgruppe
			 * 3:				Objekt:Attributgruppe:Aspekt
			 */

			//prüfen was für einen Typ das erste parts element hat
			//Debug, zweiter Durchlauf: hier hier hier: getObject liefert kein Objekt zurück -> fehler!
			quali.object = _check.getObject(parts.get(0));
			int type = MemberCheck.whatIs(quali.object);
			if(quali.object == null || (type != MemberCheck.OBJECTTYPE && type != MemberCheck.OBJECT)) {
				throw new SemanticErrorException(InterpreterErrorMessage.OBJECTTYPE_EXPECTED + parts.get(0) + parts);
			}

			if(parts.size() >= 2) {
				try	//ist notwendig, da getAttributegroup ein Exception wirft, wenn man die pid eines Aspektes angibt!
				{
					AttributeGroup atg = _model.getAttributeGroup(parts.get(1));
					quali.attributeGroup = MemberCheck.isAttributeGroupMember(quali.object, atg);
				}
				catch(ConfigurationException e) {
					throw new CriticalParserException(ErrorMessage.COMMUNICATION);
				}
				catch(Exception ex) {
					throw new ScopeException(InterpreterErrorMessage.ATTRIBUTEGROUP_MISSING);
				}
			}

			if(parts.size() >= 3) {
				if(_check.whatIs(parts.get(2)) == MemberCheck.FREE_ASPECT) {
					quali.freeAspect = parts.get(2);
				}
				else {
					Aspect asp = _model.getAspect(parts.get(2));
					MemberCheck.isAspectMember(quali.attributeGroup, asp);
					quali.aspect = asp;
				}
			}

			_pathStack.push(quali);
			_objectToScope.put(parts.get(0), quali);
		}
		catch(CriticalParserException criticalException) {
			criticalException.line = path.getLine();
			throw criticalException;
		}
		catch(ConfigurationException configurationException) {
			throw new CriticalParserException(ErrorMessage.COMMUNICATION);
		}
	}

	/**
	 * Tut nichts.
	 *
	 * @see de.bsvrz.pua.prot.interpreter.semantics.SemanticChecks#cleanUp(de.bsvrz.pua.prot.util.ProcessingParameter,boolean)
	 */
	public void cleanUp(ProcessingParameter pp, boolean processScript) throws SemanticErrorException {
	}
}
