/*
 * Copyright 2005-2008 by beck et al. projects GmbH, Munich
 * Copyright 2009 by Kappich Systemberatung, Aachen
 *
 * This file is part of de.bsvrz.pua.prot.
 *
 * de.bsvrz.pua.prot is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * de.bsvrz.pua.prot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.pua.prot.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Contact Information:
 * Kappich Systemberatung
 * Martin-Luther-Straße 14
 * 52062 Aachen, Germany
 * phone: +49 241 4090 436
 * mail: <info@kappich.de>
 */

package de.bsvrz.pua.prot.manager.requests;

import de.bsvrz.dav.daf.main.ClientDavInterface;
import de.bsvrz.dav.daf.main.Data;
import de.bsvrz.dav.daf.main.ResultData;
import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.pua.prot.manager.ManagerErrorMessage;
import de.bsvrz.pua.prot.manager.ScriptAnswer;
import de.bsvrz.pua.prot.manager.ScriptRelatedOperation;
import de.bsvrz.pua.prot.manager.datamanager.DataManagerInterface;
import de.bsvrz.pua.prot.util.ProtocolRequestResult;
import de.bsvrz.sys.funclib.debug.Debug;
import de.bsvrz.sys.funclib.losb.datk.PidScript;
import de.bsvrz.sys.funclib.losb.exceptions.FailureException;

/**
 * * Klasse die anhand eines Objekts vom Typ {@link ResultData} überprüft, welcher Skript-bezogene Auftrag ausgeführt werden soll.
 *
 * @author beck et al. projects GmbH
 * @author Martin Hilgers
 * @version $Revision$ / $Date$ / ($Author$)
 */
public class ScriptRequest implements DataManagerRequest {

	/** Auftragsparameter */
	private ResultData _resultData[];

	/** Debug - Ausgabe */
	protected static final Debug debug = Debug.getLogger();

	/** Verbindung zum Datenverteiler */
	protected ClientDavInterface dav;

	/** Auftraggeber, and den das Ergebnis gesendet wird */
	protected SystemObject client;

	/** Anfrage-Id */
	protected long requestId;

	/** Skriptobjekt */
	protected SystemObject script;

	/** Schnittstelle zum DataManager */
	protected DataManagerInterface dataManager;

	/** Zeigt an das der AUftrag abgebrochen werden soll. */
	private boolean _abort;

	/** Gibt an ob der Pretty Printer verwendet werden soll. */
	private boolean _prettyPrinter;

	/**
	 * Erzeugt das Objekt. {@link #client}, {@link #requestId} und {@link #script} werden mit den Werten aus <code>resultData[0]</code> initialisiert!
	 *
	 * @param dav           Verbindung zum Datenverteiler.
	 * @param dataManager   Submodul Daten Manager.
	 * @param resultData    Aufrufparameter.
	 * @param prettyPrinter Gibt an ob der Pretty Printer verwendet werden soll.
	 */
	public ScriptRequest(ClientDavInterface dav, DataManagerInterface dataManager, ResultData resultData[], boolean prettyPrinter) {
		this.dav = dav;
		this._resultData = resultData;
		this.dataManager = dataManager;
		this._prettyPrinter = prettyPrinter;

		if(resultData.length > 0 && resultData[0].getData() != null) {
			script = resultData[0].getObject();
			client = resultData[0].getData().getItem(PidScript.atlExtendedInformation).getReferenceValue(PidScript.sender).getSystemObject();
			requestId = resultData[0].getData().getItem(PidScript.atlExtendedInformation).getUnscaledValue(PidScript.requestId).longValue();
		}
	}

	/**
	 * Bildet Aufträge aus den Auftragsparametern und führt diese aus. Mögliche Aufträge sind<br> - Skriptquelltext zurückliefern<br> - Skript anlegen / editieren
	 *
	 * @return Liefert immer eine positive Rückmeldung.
	 */
	public ProtocolRequestResult process() {
		for(ResultData rd : _resultData) {
			if(isAborted()) break;
			Data data = rd.getData();
			if(data != null) {
				//TODO Genauere Fehlermeldung, wenn diverse Felder nicht gesetzt wurden!
				//Auftraggeber, Anfrage-Id, Operationscode und Skriptobjekt auslesen
				// BZW was passiert, wenn es hier eine nullpointer-exception gibt, weil sich die pids geändert haben!
				Data atl = data.getItem(PidScript.atlExtendedInformation);
				SystemObject clientObject = atl.getReferenceValue(PidScript.sender).getSystemObject();
				long id = atl.getUnscaledValue(PidScript.requestId).longValue();
				int opCode = atl.getUnscaledValue(PidScript.operationCode).intValue();
				SystemObject scriptObject = rd.getObject();
				switch(opCode) {
					case ScriptRelatedOperation.CREATE:
						CreateScriptRequest csr = new CreateScriptRequest(dav, dataManager, rd, _prettyPrinter);
						csr.process();
						break;
					case ScriptRelatedOperation.SOURCE:
						RetrieveScriptSourceRequest rsr = new RetrieveScriptSourceRequest(dav, dataManager, rd);
						rsr.process();
						break;
					case ScriptRelatedOperation.DEFAULTS:
						RetrieveScriptDefaultsRequest rdr = new RetrieveScriptDefaultsRequest(dav, dataManager, rd);
						rdr.process();
						break;
					default:
						try {
							ScriptAnswer.sendError(dav, scriptObject, clientObject, id, ManagerErrorMessage.INVALID_OPCODE + opCode);
						}
						catch(FailureException e) {
							debug.warning("Fehler beim Senden einer Fehlermeldung.", e);
						}
				}
			}
		}
		return new ProtocolRequestResult();
	}

	/** @see de.bsvrz.pua.prot.manager.requests.DataManagerRequest#abort() */
	public void abort() {
		this._abort = true;
	}

	/** @return <code>true</code> wenn der Auftrag abgebrochen werden soll. */
	public boolean isAborted() {
		return _abort;
	}

	/** @return <code>true</code> falls der Pretty Printer verwendet werden soll.<code>false</code>sonst. */
	public boolean printPretty() {
		return _prettyPrinter;
	}
}
