/*
 * Copyright 2005-2008 by beck et al. projects GmbH, Munich
 * Copyright 2009 by Kappich Systemberatung, Aachen
 *
 * This file is part of de.bsvrz.pua.prot.
 *
 * de.bsvrz.pua.prot is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * de.bsvrz.pua.prot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.pua.prot.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Contact Information:
 * Kappich Systemberatung
 * Martin-Luther-Straße 14
 * 52062 Aachen, Germany
 * phone: +49 241 4090 436
 * mail: <info@kappich.de>
 */

package de.bsvrz.pua.prot.processing.archivebuffer;

import de.bsvrz.dav.daf.main.DataDescription;
import de.bsvrz.dav.daf.main.DataState;
import de.bsvrz.dav.daf.main.archive.ArchiveData;
import de.bsvrz.dav.daf.main.archive.ArchiveDataKind;
import de.bsvrz.dav.daf.main.archive.ArchiveDataStream;
import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.pua.prot.processing.ProcessingErrorMessage;
import de.bsvrz.pua.prot.processing.ValueProvider;
import de.bsvrz.pua.prot.processing.onlinebuffer.ExtResultData;
import de.bsvrz.sys.funclib.losb.datk.ProtocolType;
import de.bsvrz.sys.funclib.debug.Debug;
import de.bsvrz.sys.funclib.losb.exceptions.FailureException;
import de.bsvrz.sys.funclib.losb.exceptions.LoggerException;
import de.bsvrz.sys.funclib.losb.util.Util;

import java.io.IOException;

/**
 * Klasse die Methoden zur Bestimmung von Archiv-Werten für reale Attribute und Attributgruppen durchführt. Basisklasse für Attributgruppen oder Reales
 * Attribute mit zusätzlichen Informationen für die Datenaufbereitung.
 *
 * @author beck et al. projects GmbH
 * @author Martin Hilgers
 * @version $Revision$ / $Date$ / ($Author$)
 */
public abstract class RealArchiveElement extends ValueProvider {

	/** Anzufragende Attributgruppe & Aspekt */
	private DataDescription _dataDescription;

	/** Anzufragendes Objekt. */
	private SystemObject _systemObject;

	/** Verbindung zum Archivsystem */
	private ArchiveUser _archiveUser;

	/** Zeigt an ob noch weitere Daten geliefert werden. */
	private boolean _hasNext = true;

	/** Debug - Ausgabe */
	private static final Debug _debug = Debug.getLogger();

	/**
	 * Konstruktor.
	 *
	 * @param archiveUser     Verbindung zum Archivsystem
	 * @param dataDescription Datenbeschreibung. Wichtig: Simulationsvariante muss explizit gesetzt werden! Sonst liefert <code>dataDescription.getSimVar()</code>
	 *                        ggf. -1, und das kann vom Archivsystem nicht verarbeitet werden.
	 * @param object          Objekt für das die Archivanfrage gestellt werden soll.
	 * @param protocolType    Protokollart
	 * @param isShown         Wird das Element im Spalten-Bereich verwendet? <code>true</code>Element wird im Spalten Bereich verwendet.
	 * @param padData         <code>true</code> falls Daten aufgefüllt werden dürfen.
	 */
	public RealArchiveElement(
			ArchiveUser archiveUser, DataDescription dataDescription, SystemObject object, ProtocolType protocolType, boolean isShown, boolean padData) {
		super(protocolType, isShown, padData);
		if(object == null) throw new NullPointerException();

		this._archiveUser = archiveUser;
		this._dataDescription = dataDescription;
		this._systemObject = object;
	}

	/**
	 * Führt eine Archivanfrage durch. Dabei werden die im Konstruktor übergebenen Standardparameter von <code>archiveUser</code>verwendet. Nach dem erfolgreichen
	 * Ausführen dieser Methode wird {@link #hasNext()} <code>true</code> zurückliefern.
	 *
	 * @throws FailureException     Fehler bei der Archivanfrage
	 * @throws InterruptedException Warten auf Antwortdatensatz wurde unterbrochen.
	 */
	public void request() throws FailureException, InterruptedException {
		try {
			_archiveUser.request(this, _dataDescription, _systemObject);
			_hasNext = true;
			//Es wurde ein neues Intervall begonnen -> falls noch alte Daten vorliegen, dürfen diese nicht zum auffüllen verwendet werden
			reset();
		}
		catch(IllegalStateException e) {
			throw new FailureException(
					"RealArchiveElement.request(): " + ProcessingErrorMessage.ARCHIVE_COMMUNICATION + e.getMessage(), e, LoggerException.WARNING
			);
		}
	}

	/**
	 * Holt den nächsten Datensatz aus dem Archivsystem.
	 *
	 * @param timeout Hat keine Bedeutung
	 *
	 * @return <code>true</code> falls Daten empfangen wurde, <code>false</code> falls es keine Daten gibt.
	 *
	 * @throws FailureException     Fehler bei der Archivanfrage
	 * @throws InterruptedException Warten auf Antwortdatensatz wurde unterbrochen.
	 */
	@Override
    protected boolean executeTake(long timeout) throws FailureException, InterruptedException {
		try {
			ArchiveData ad = _archiveUser.take(this);
			if(ad == null) {
				_debug.fine("Es liegen keine weiteren Daten mehr vor für ", this);
				_hasNext = false;	//Thread läuft hier rein -> setzt hasnext auf false -> Processing.process() Hauptschleife bricht ab..
				return false;
			}
			else {
				value = new ExtResultData(ad, ad.getData(), ad.getDataIndex(), ad.getArchiveTime());

				if(ad.getDataType() == DataState.END_OF_ARCHIVE){
					value.setDataTime(_archiveUser.getTimeSpec().getIntervalEnd());
				}

				_debug.fine("Daten abgeholt: ", value);
				return true;
			}
		}
		catch(IllegalStateException e) {
			throw new FailureException(ProcessingErrorMessage.ARCHIVE_REQUEST_ABORTED + e.getMessage(), e, LoggerException.WARNING);
		}
		catch(IOException e) {
			throw new FailureException(
					"RealArchiveElement.take(): " + ProcessingErrorMessage.ARCHIVE_COMMUNICATION + e.getMessage(), e, LoggerException.WARNING
			);
		}
	}

	/** @see de.bsvrz.pua.prot.processing.ValueProvider#updateDataInformation() */
	@Override
    public boolean updateDataInformation() {
		if(value != null) {
			dataInformation = dataInformation.createDataInformation(value, isTimeSorted(value));
			return dataInformation.insertEmpty();
		}
		else {
			return false;
		}
	}

	/**
	 * Zeigt ob die Daten nach Datenzeitstempel geordnet sind oder nicht.
	 *
	 * @param data Datensatz vom Archivsystem. Darf <code>null</code> sein; dann wird <code>false</code> zurückgeliefert.
	 *
	 * @return	<code>true</code> falls die Daten nach Datenzeitstempel sortiert sind. <code>false</code> sonst.
	 */
	private boolean isTimeSorted(ExtResultData data) {
		if(data == null) {
			return false;
		}
		else
		if(data.getDataKind().equals(ArchiveDataKind.REQUESTED) || data.getDataKind().equals(ArchiveDataKind.REQUESTED_DELAYED) || data.getDataKind().equals(
				ArchiveDataKind.ONLINE_DELAYED
		)) {
			return true;
		}
		else {
			return false;
		}
	}


	/** @see de.bsvrz.pua.prot.processing.ValueProvider#hasNext() */
	@Override
    public boolean hasNext() {
		return _hasNext;
	}

	/**
	 * Liefert eine Verbindung zum Archivsystem mit Standardparametern für die Archivanfrage.
	 *
	 * @return Liefert eine Verbindung zum Archivsystem
	 */
	protected ArchiveUser getArchiveUser() {
		return _archiveUser;
	}

	/**
	 * Bricht die Datensammlung ab.
	 *
	 * @see ArchiveDataStream#abort()
	 */
	@Override
    public void abort() {
		if(_archiveUser != null) _archiveUser.abort();
	}

	/** @see java.lang.Object#toString() */
	@Override
    public String toString() {
		StringBuilder sb = new StringBuilder();
		if(value != null) sb.append(Util.msToDate(value.getDataTime()) + ", " + Util.dIdx2Str(value.getDataIndex()) + ": ");
		sb.append(_systemObject + ":" + _dataDescription.toString());
		return sb.toString();
	}
}
