/*
 * Copyright 2005-2008 by beck et al. projects GmbH, Munich
 * Copyright 2009 by Kappich Systemberatung, Aachen
 *
 * This file is part of de.bsvrz.pua.prot.
 *
 * de.bsvrz.pua.prot is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * de.bsvrz.pua.prot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.pua.prot.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Contact Information:
 * Kappich Systemberatung
 * Martin-Luther-Straße 14
 * 52062 Aachen, Germany
 * phone: +49 241 4090 436
 * mail: <info@kappich.de>
 */

package de.bsvrz.pua.prot.util;

import de.bsvrz.dav.daf.main.archive.ArchiveDataKind;
import de.bsvrz.dav.daf.main.archive.ArchiveDataKindCombination;
import de.bsvrz.dav.daf.main.config.ClientApplication;
import de.bsvrz.dav.daf.main.config.DynamicObject;
import de.bsvrz.dav.daf.main.config.ObjectLookup;
import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.pua.prot.interpreter.InterpreterErrorMessage;
import de.bsvrz.sys.funclib.kappich.annotations.Nullable;
import de.bsvrz.sys.funclib.losb.datk.NoChangeMarker;
import de.bsvrz.sys.funclib.losb.datk.ProtocolType;
import de.bsvrz.sys.funclib.losb.exceptions.FailureException;
import de.bsvrz.sys.funclib.losb.exceptions.LoggerException;
import de.bsvrz.sys.funclib.losb.messages.ErrorMessage;
import de.bsvrz.sys.funclib.losb.util.Tuple;
import de.bsvrz.sys.funclib.losb.util.Util;

import java.io.*;
import java.util.*;
import java.util.stream.Collectors;

/**
 * Datenklasse, die alle Parameter, die für die Protokollerstellung angegeben müssen, zusammenfasst
 *
 * @author beck et al. projects GmbH
 * @author Martin Hilgers
 * @version $Revision$ / $Date$ / ($Author$)
 */
@SuppressWarnings({"IOResourceOpenedButNotSafelyClosed"})
public final class ProcessingParameter {

	/**
	 * Objekt, das das Skript darstellt. Ein null-Wert zeigt eine fehlerhafte Initialisierung der Klasse über den deprecated Konstruktor ProcessingParameter() an
	 * und führt bei Anfragen zu einer NullPointerException auf der Clientseite
	 */
	private SystemObject _script = null;

	/** Liste mit Zeitbereichen (Tuple<Start-Ende>) */
	private List<Tuple<Long, Long>> _periodList = new ArrayList<>();

	/** Hauptobjekt. Darf null sein. Wenn null wird gegebenenfalls ein im Skript definiertes Standardobjekt verwendet. */
	private SystemObject _mainObject = null;

	/** Simulationsvariante */
	private short _simVar = Qualifier.INVALID_SIMVAR;

	/** Map mit Aspekt-Bindungen (Zuordnung Alias->Aspekt) */
	private HashMap<String, String> _aspBindings = new HashMap<>();

	/** Map mit Pseudoobjekt-Definitionen (Zuordnung Alias->Objekte) */
	private HashMap<String, List<String>> _pseudoObjects = new HashMap<>();

	/** Ob das Protokoll gespeichert werden soll */
	private boolean _saveProtocol = false;

	/**
	 * Typ des Protokolls. Immer != null. Bei Undefined wird der Standardwert im Skript verwendet. Ist im Script nichts angegeben wird ein Zustandsprotokoll erzeugt.
	 */
	private ProtocolType _protocolType = ProtocolType.Undefined;

	/**
	 * Markierungsart für "Keine Änderungen"-Informationen. Entweder Zeilenweise oder pro einzelnem Datensatz.
	 */
	private NoChangeMarker _noChangeMarker = NoChangeMarker.Undefined;

	/** Set mit Datenarten */
	private Set<ArchiveDataKind> _archiveDataKind = new HashSet<>();

	/** Infotext der beim Protokollerstellen angegeben werden kann. null-Werte sind ungültig. */
	private String _infoText = "";

	/** Name des Erstellers des Protokolls (Freitext). null-Werte sind ungültig. */
	private String _creatorName = "";

	/** Erstellungsdatum des Protokolls in ms seit 1.Januar 1970. Wird standardmäßig im Konstruktor auf System.currentTimeMillis() gesetzt. */
	private long _creationDate = System.currentTimeMillis();

	/** Gibt an, ob das Protokoll öffentlich sein soll */
	private boolean _isPublic = true;
	
	/** Erstellender Benutzer. Anders als {@link #_creatorName} wird dieser Wert für die Rechteprüfung verwendet. */
	@Nullable
	private DynamicObject _creatorUser = null;

	/**
	 * Erzeugt Auftragsparameter mit Simulationsvariante 0.
	 *
	 * @param script     Skript das verwendet werden soll.
	 *
	 * @throws FailureException Ungültige Parameter.
	 */
	public ProcessingParameter(final SystemObject script) throws FailureException {
		this.setScript(script);
		this.setSimVar((short)0);
		_archiveDataKind.add(ArchiveDataKind.ONLINE);
	}

	/**
	 * Erzeugt Auftragsparameter.
	 *
	 * @param script     Skript das verwendet werden soll.
	 * @param mainObject Hauptobjekt (kann null sein)
	 * @param simVar     Simulationsvariante
	 *
	 * @throws FailureException Ungültige Parameter.
	 */
	public ProcessingParameter(final SystemObject script, final SystemObject mainObject, final short simVar) throws FailureException {
		this.setScript(script);
		this.setMainObject(mainObject);
		this.setSimVar(simVar);

		_archiveDataKind.add(ArchiveDataKind.ONLINE);
	}

	/**
	 * Erzeugt Auftragsparameter.
	 *
	 * @param script     Skript das verwendet werden soll.
	 * @param mainObject Hauptobjekt (kann null sein)
	 * @param simVar     Simulationsvariante
	 * @param start      Startzeitpunkt des Protokolls.
	 * @param stop       Endzeitpunkt des Protokolls.
	 *
	 * @throws FailureException Ungültige Parameter.
	 */
	public ProcessingParameter(final SystemObject script, final SystemObject mainObject, final short simVar, final long start, final long stop)
			throws FailureException {
		this(script, mainObject, simVar);

		addPeriod(new Tuple<>(start, stop));
	}


	/**
	 * Erzeugt Auftragsparameter.
	 *
	 * @param script     Skript das verwendet werden soll.
	 * @param mainObject Hauptobjekt (kann null sein)
	 * @param simVar     Simulationsvariante
	 * @param periods    Zeitintervalle des Protokolls.
	 *
	 * @throws FailureException Ungültige Parameter.
	 */
	@Deprecated
	public ProcessingParameter(final SystemObject script, final SystemObject mainObject, final short simVar, final Collection<Tuple<Long, Long>> periods)
			throws FailureException {
		this(script, mainObject, simVar);
		setPeriodList(periods);
	}

	/**
	 * Erzeugt Auftragsparameter.
	 *
	 * @param script          Skript das verwendet werden soll.
	 * @param mainObject      Hauptobjekt (kann null sein)
	 * @param simVar          Simulationsvariante
	 * @param periods         Zeitintervalle des Protokolls.
	 * @param aspBindings     Aspekt-Bindungen oder {@link Collections#EMPTY_MAP} falls keine gewünscht sind.
	 * @param pseudoObjects   Pseudo-Objekt-Definitionen oder {@link Collections#EMPTY_MAP} falls keine gewünscht sind.
	 * @param saveProtocol    Ob das Protokoll gespeichert werden soll
	 * @param deltaProtocol   Ob das Protokoll ein Änderungsprotokoll ist (true: Ja, false: Zustandsprotokoll, null: Standardwert im Skript verwenden)
	 * @param archiveDataKind Liste mit Datensatzarten für die das Protokoll erstellt werden soll
	 * @param creatorName     Name des Erstellers
	 * @param isPublic        Ist das Protokoll öffentlich?
	 * @param infoText        Beschreibungstext
	 *
	 * @throws FailureException Ungültige Parameter.
	 */
	@Deprecated
	public ProcessingParameter(
			final SystemObject script,
			final SystemObject mainObject,
			final short simVar,
			final Collection<Tuple<Long, Long>> periods,
			final Map<String, String> aspBindings,
			final Map<String, List<String>> pseudoObjects,
			final boolean saveProtocol,
			final Boolean deltaProtocol,
			final Collection<ArchiveDataKind> archiveDataKind,
			final String creatorName,
			final boolean isPublic,
			final String infoText) throws FailureException {
		this(script, mainObject, simVar, periods);
		setAspBindings(aspBindings);
		setPseudoObjects2(pseudoObjects);
		setSaveProtocol(saveProtocol);
		setDeltaProtocol(deltaProtocol);
		setArchiveDataKind(archiveDataKind);
		setCreatorName(creatorName);
		setIsPublic(isPublic);
		setInfoText(infoText);
	}

	/**
	 * Erzeugt Auftragsparameter.
	 *
	 * @param script          Skript das verwendet werden soll.
	 * @param mainObject      Hauptobjekt (kann null sein)
	 * @param simVar          Simulationsvariante
	 * @param periods         Zeitintervalle des Protokolls.
	 * @param aspBindings     Aspekt-Bindungen oder {@link java.util.Collections#EMPTY_MAP} falls keine gewünscht sind.
	 * @param pseudoObjects   Pseudo-Objekt-Definitionen oder {@link java.util.Collections#EMPTY_MAP} falls keine gewünscht sind.
	 * @param saveProtocol    Ob das Protokoll gespeichert werden soll
	 * @param protocolType    Zu erstellende Protokollart. <code>null</code> wird als {@link de.bsvrz.sys.funclib.losb.datk.ProtocolType#Undefined} interpretiert.
	 * @param noChangeMarker  Markierungsart für "Keine Änderungen" bei Zustandsprotokollen, entweder Zeilenweise oder für jede einzelne Zelle/jeden einzelnen Datensatz. <code>null</code> wird als {@link de.bsvrz.sys.funclib.losb.datk.NoChangeMarker#Undefined} interpretiert.
	 * @param archiveDataKind Liste mit Datensatzarten für die das Protokoll erstellt werden soll
	 * @param creatorName     Name des Erstellers
	 * @param isPublic        Ist das Protokoll öffentlich?
	 * @param infoText        Beschreibungstext
	 *
	 * @throws FailureException Ungültige Parameter.
	 */
	@Deprecated
	public ProcessingParameter(
			final SystemObject script,
			final SystemObject mainObject,
			final short simVar,
			final Collection<Tuple<Long, Long>> periods,
			final Map<String, String> aspBindings,
			final Map<String, List<String>> pseudoObjects,
			final boolean saveProtocol,
			final ProtocolType protocolType,
			final NoChangeMarker noChangeMarker,
			final Collection<ArchiveDataKind> archiveDataKind,
			final String creatorName,
			final boolean isPublic,
			final String infoText) throws FailureException {
		this(script, mainObject, simVar, periods);
		setAspBindings(aspBindings);
		setPseudoObjects2(pseudoObjects);
		setSaveProtocol(saveProtocol);
		setProtocolType(protocolType);
		setNoChangeMarker(noChangeMarker);
		setArchiveDataKind(archiveDataKind);
		setCreatorName(creatorName);
		setIsPublic(isPublic);
		setInfoText(infoText);
	}

	/**
	 * Erzeugt Auftragsparameter.
	 *
	 * @param script     Skript das verwendet werden soll.
	 * @param mainObject Hauptobjekt (kann null sein)
	 * @param simVar     Simulationsvariante
	 * @param periods    Zeitintervalle des Protokolls.
	 *
	 * @throws FailureException Ungültige Parameter.
	 */
	public ProcessingParameter(final SystemObject script, final SystemObject mainObject, final short simVar, final PeriodSet periods)
			throws FailureException {
		this(script, mainObject, simVar);
		setPeriodSet(periods);
	}

	/**
	 * Erzeugt Auftragsparameter.
	 *
	 * @param script          Skript das verwendet werden soll.
	 * @param mainObject      Hauptobjekt (kann null sein)
	 * @param simVar          Simulationsvariante
	 * @param periods         Zeitintervalle des Protokolls.
	 * @param aspBindings     Aspekt-Bindungen oder {@link Collections#EMPTY_MAP} falls keine gewünscht sind.
	 * @param pseudoObjects   Pseudo-Objekt-Definitionen oder {@link Collections#EMPTY_MAP} falls keine gewünscht sind.
	 * @param saveProtocol    Ob das Protokoll gespeichert werden soll
	 * @param protocolType    Zu erstellende Protokollart. <code>null</code> wird als {@link ProtocolType#Undefined} interpretiert.
	 * @param noChangeMarker  Markierungsart für "Keine Änderungen" bei Zustandsprotokollen, entweder Zeilenweise oder für jede einzelne Zelle/jeden einzelnen Datensatz. <code>null</code> wird als {@link NoChangeMarker#Undefined} interpretiert.
	 * @param archiveDataKind Liste mit Datensatzarten für die das Protokoll erstellt werden soll
	 * @param creatorName     Name des Erstellers
	 * @param isPublic        Ist das Protokoll öffentlich?
	 * @param infoText        Beschreibungstext
	 *
	 * @throws FailureException Ungültige Parameter.
	 */
	public ProcessingParameter(
			final SystemObject script,
			final SystemObject mainObject,
			final short simVar,
			final PeriodSet periods,
			final Map<String, String> aspBindings,
			final Map<String, List<String>> pseudoObjects,
			final boolean saveProtocol,
			final ProtocolType protocolType,
			final NoChangeMarker noChangeMarker,
			final Collection<ArchiveDataKind> archiveDataKind,
			final String creatorName,
			final boolean isPublic,
			final String infoText) throws FailureException {
		this(script, mainObject, simVar, periods);
		setAspBindings(aspBindings);
		setPseudoObjects2(pseudoObjects);
		setSaveProtocol(saveProtocol);
		setProtocolType(protocolType);
		setNoChangeMarker(noChangeMarker);
		setArchiveDataKind(archiveDataKind);
		setCreatorName(creatorName);
		setIsPublic(isPublic);
		setInfoText(infoText);
	}

	/**
	 * Erstellt ein neues ProcessingParameter-Objekt. Nach dem Erstellen muss zwingend {@link #setScript(de.bsvrz.dav.daf.main.config.SystemObject)} und {{@link
	 * #setSimVar(short)}} aufgerufen werden.
	 *
	 * @deprecated Bei der Benutzung der anderen Konstruktoren enthält das Objekt in jedem Fall gültige Werte, deren Benutzung ist daher zu empfehlen.
	 */
	@Deprecated
	public ProcessingParameter() {
	}

	/**
	 * Serialisiert das Objekt in einen Byte-Array.
	 *
	 * @return Byte Array.
	 *
	 * @throws FailureException Fehler beim Serialisieren.
	 */
	public byte[] serialize() throws FailureException {
		ObjectOutputStream oos = null;
		try {
			//Daten serialisieren:
			final ByteArrayOutputStream bos = new ByteArrayOutputStream();
			oos = new ObjectOutputStream(bos);
			//archiveDataKind
			final ArrayList<Long> list = new ArrayList<>();
			for(final ArchiveDataKind adk : getArchiveDataKind()) {
				if(adk.equals(ArchiveDataKind.ONLINE)) {
					list.add(0L);
				}
				else if(adk.equals(ArchiveDataKind.ONLINE_DELAYED)) {
					list.add(1L);
				}
				else if(adk.equals(ArchiveDataKind.REQUESTED)) {
					list.add(2L);
				}
				else if(adk.equals(ArchiveDataKind.REQUESTED_DELAYED)) {
					list.add(3L);
				}
			}
			oos.writeObject(list);
			//mainObject
			final SystemObject mainObject = getMainObject();
			if(mainObject != null) {
				oos.writeLong(mainObject.getId());
			}
			else {
				oos.writeLong(0);
			}
			//script
			oos.writeLong(getScript().getId());

			oos.writeObject(getAspBindings());
			oos.writeObject(getDeltaProtocol()); // Protokollart für ältere PuA-Versionen
			oos.writeObject(getPeriodList());
			oos.writeObject(getPseudoObjects2());
			oos.writeBoolean(getSaveProtocol());
			oos.writeShort(getSimVar());

			oos.writeObject(getInfoText());
			oos.writeObject(getCreatorName());
			oos.writeLong(getCreationDate());
			oos.writeBoolean(isPublic());

			// Erweiterter Enum: Protokollart
			oos.writeObject(getProtocolType().name());

			// Markierungsart "Keine Änderungen"-Datensätze
			oos.writeObject(getNoChangeMarker().name());

			DynamicObject creatorUser = getCreatorUser();
			if(creatorUser != null) {
				oos.writeLong(creatorUser.getId());
			}
			else {
				oos.writeLong(0);
			}

			oos.flush();

			return bos.toByteArray();
		}
		catch(IOException e) {
			throw new FailureException(ErrorMessage.CAN_NOT_SERIALIZE, e, LoggerException.WARNING);
		}
		finally {
			Util.close(oos);
		}
	}

	/**
	 * Deserialisiert ein Objekt aus einem Byte-Array. Der übliche Java Serialisierer kann nicht verwendet werden, da Objekte der Klasse <code>SystemObject</code>
	 * damit nicht serialisiert werden können.
	 *
	 * @param model Verbindung zur Konfiguration.
	 * @param ba    Serialisiertes Objekt als Byte-Array.
	 *
	 * @return Deserialisiertes Objekt.
	 *
	 * @throws FailureException Fehler beim Deserialisieren.
	 */
	@SuppressWarnings({"unchecked", "deprecation"})
	public static ProcessingParameter deserialize(final ObjectLookup model, final byte[] ba) throws FailureException {
		ObjectInputStream ois = null;
		try {
			final ByteArrayInputStream bis = new ByteArrayInputStream(ba);
			ois = new ObjectInputStream(bis);
			final ProcessingParameter pp = new ProcessingParameter();
			//archiveDataKind
			final Collection<Long> list = (Collection<Long>)ois.readObject();
			final Set<ArchiveDataKind> archiveDataKindList = new HashSet<>(list.size());
			for(final long item : list) {
				if(item == 0L) {
					archiveDataKindList.add(ArchiveDataKind.ONLINE);
				}
				else if(item == 1L) {
					archiveDataKindList.add(ArchiveDataKind.ONLINE_DELAYED);
				}
				else if(item == 2L) {
					archiveDataKindList.add(ArchiveDataKind.REQUESTED);
				}
				else if(item == 3L) {
					archiveDataKindList.add(ArchiveDataKind.REQUESTED_DELAYED);
				}
			}
			pp.setArchiveDataKind(archiveDataKindList);
			//mainObject
			final long mainObjectId = ois.readLong();
			if(mainObjectId != 0) {
				pp.setMainObject(model.getObject(mainObjectId));
			}
			else {
				pp.setMainObject(null);
			}
			//script
			pp.setScript(model.getObject(ois.readLong()));

			pp.setAspBindings((Map<String, String>) ois.readObject());
			pp.setDeltaProtocol((Boolean) ois.readObject());
			pp.setPeriodList((List<Tuple<Long, Long>>) ois.readObject());
			pp.setPseudoObjects2((Map<String, List<String>>) ois.readObject());
			pp.setSaveProtocol(ois.readBoolean());
			pp.setSimVar(ois.readShort());

			pp.setInfoText((String) ois.readObject());
			pp.setCreatorName((String) ois.readObject());
			pp.setCreationDate(ois.readLong());
			pp.setIsPublic(ois.readBoolean());

			try {
				pp.setProtocolType(ProtocolType.valueOf((String) ois.readObject()));
				pp.setNoChangeMarker(NoChangeMarker.valueOf((String) ois.readObject()));

				try {
					SystemObject userObject = model.getObject(ois.readLong());
					if(userObject instanceof DynamicObject) {
						pp._creatorUser = (DynamicObject) userObject;
					}
				}
				catch(IllegalArgumentException ignored){
					// Unbekannte Protokolltypen ignorieren
				}
				catch(EOFException ignored){
					// Alte Client serialisieren die erweiterten Enums nicht, also einfach Standardwerte annehmen, bzw. im Falle
					// der Protokollart den vorher eingelesenen Wert über setDeltaProtocol()
				}
				
			}
			catch(IllegalArgumentException ignored){
				// Unbekannte Protokolltypen ignorieren
			}
			catch(EOFException ignored){
				// Alte Client serialisieren die erweiterten Enums nicht, also einfach Standardwerte annehmen, bzw. im Falle
				// der Protokollart den vorher eingelesenen Wert über setDeltaProtocol()
			}

			return pp;
		}
		catch(Exception e) {
			e.printStackTrace();
			throw new FailureException(ErrorMessage.CAN_NOT_DESERIALIZE, e, LoggerException.WARNING);
		}
		finally {
			Util.close(ois);
		}
	}

	/** @see java.lang.Object#toString() */
	@Override
	public String toString() {
		final StringBuffer sb = new StringBuffer();
		sb.append("Archivdatenart: ");
		for(final ArchiveDataKind adk : getArchiveDataKind()) {
			sb.append(adk).append(", ");
		}

		sb.append("Bindungen: ");
		for(final String key : getAspBindings().keySet()) {
			sb.append(key).append("->").append(getAspBindings().get(key)).append(", ");
		}

		switch(_protocolType){
			case StatusProtocol:
				sb.append("Protokollart: Statusprotokoll, ");
				break;
			case DeltaProtocol:
				sb.append("Protokollart: Änderungsprotokoll, ");
				break;
			case EventProtocol:
				sb.append("Protokollart: Ereignisprotokoll, ");
				break;
			case Undefined:
				break;
		}

		switch(_noChangeMarker){
			case Cell:
				sb.append("Unverändertkennung: Pro Zelle, ");
				break;
			case Row:
				sb.append("Unverändertkennung: Pro Zeile, ");
				break;
			case Undefined:
				break;
		}

		sb.append("Hauptobjekt: ").append(getMainObject().getPid()).append(", ");

		sb.append("Intervalle : ");
		for(final Tuple<Long, Long> period : getPeriodList()) {
			Util.msToDate(sb, period.first);
			sb.append(" - ");
			Util.msToDate(sb, period.last);
			sb.append(", ");
		}

		sb.append("Pseudoobjektdefinitionen: ");
		for(final String alias : getPseudoObjects2().keySet()) {
			sb.append(alias).append("->").append(getPseudoObjects2().get(alias)).append(", ");
		}

		sb.append("Speichern: ").append((getSaveProtocol()) ? "ja, " : "nein, ");

		sb.append("Skript: ").append(getScript().getNameOrPidOrId()).append(", ");

		sb.append("Simulationsvariante: ").append(getSimVar());

		sb.append(", Erstellt: ").append(new Date(getCreationDate()));

		sb.append(" von: ").append(getCreatorName());

		DynamicObject creatorUser = getCreatorUser();
		if(creatorUser != null) {
			sb.append(" (").append(creatorUser).append(")");
		}

		sb.append(", Beschreibung: ").append(getInfoText());

		return sb.toString();
	}

	/**
	 * Gibt das Skriptobjekt zurück
	 *
	 * @return Skriptobjekt (nicht null)
	 */
	public SystemObject getScript() {
		return _script;
	}

	/**
	 * Setzt das Skriptobjekt
	 *
	 * @param script Skriptobjekt (null ist ungültig)
	 *
	 * @throws FailureException bei null-parameter
	 */
	public void setScript(final SystemObject script) throws FailureException {
		if(script == null) throw new FailureException(ErrorMessage.PARAM_IS_NULL + "script", LoggerException.OTHER);
		this._script = script;
	}

	/**
	 * Gibt Zeitbereiche, die berücksichtigt werden sollen, zurück
	 *
	 * @return Zeitbereiche
	 */
	public List<Tuple<Long, Long>> getPeriodList() {
		return Collections.unmodifiableList(_periodList);
	}

	/**
	 * Setzt Zeitbereiche
	 *
	 * @param periodList Liste mit Zeitbereichen (Tuple&lt;Start,Ende&gt;)
	 *
	 * @throws FailureException Fehlerhafter Parameter
	 */
	public void setPeriodList(final Collection<Tuple<Long, Long>> periodList) throws FailureException {
		if(periodList == null) throw new FailureException(ErrorMessage.PARAM_IS_NULL + "periodList", LoggerException.OTHER);
		for(final Tuple<Long, Long> period : periodList) {
			validatePeriod(period);
		}
		this._periodList = new ArrayList<>(periodList);
	}
	
	/**
	 * Gibt Zeitbereiche, die berücksichtigt werden sollen, zurück
	 *
	 * @return Zeitbereiche
	 */
	public PeriodSet getPeriodSet() {
		return PeriodSet.copyOf(_periodList.stream().map(it -> DiscretePeriod.of(it.first, it.last)).collect(Collectors.toList()));
	}

	/**
	 * Setzt Zeitbereiche
	 *
	 * @param periodSet Zeitbereiche
	 */
	public void setPeriodSet(final PeriodSet periodSet) throws FailureException {
		if(periodSet == null) throw new FailureException(ErrorMessage.PARAM_IS_NULL + "periodSet", LoggerException.OTHER);
		this._periodList = periodSet.getPeriods().stream().map(it -> new Tuple<>(it.getFrom().toEpochMilli(), it.getTo().toEpochMilli())).collect(Collectors.toList());
	}

	private void validatePeriod(final Tuple<Long, Long> period) throws FailureException {
		if(period.first > period.last) throw new FailureException(ErrorMessage.INVALID_PERIOD, LoggerException.OTHER);
	}

	/**
	 * Fügt einen Zeitbereich hinzu
	 *
	 * @param period Zeitbereich
	 *
	 * @throws FailureException Fehlerhafter Parameter
	 */
	public void addPeriod(final Tuple<Long, Long> period) throws FailureException {
		if(period == null) throw new FailureException(ErrorMessage.PARAM_IS_NULL + "period", LoggerException.OTHER);
		this._periodList.add(period);
	}

	/**
	 * Entfernt einen Zeitbereich
	 *
	 * @param period Zeitbereich
	 *
	 * @throws FailureException Fehlerhafter Parameter
	 */
	public void removePeriod(final Tuple<Long, Long> period) throws FailureException {
		if(period == null) throw new FailureException(ErrorMessage.PARAM_IS_NULL + "period", LoggerException.OTHER);
		this._periodList.remove(period);
	}

	/**
	 * Gibt das Hauptobjekt des Protokolls zurück
	 *
	 * @return Hauptobjekt oder null falls keines definiert wurde.
	 */
	public SystemObject getMainObject() {
		return _mainObject;
	}

	/**
	 * Setzt das Hauptobjekt
	 *
	 * @param mainObject Hauptobjekt oder null falls keines definiert sein soll.
	 */
	public void setMainObject(final SystemObject mainObject) {
		this._mainObject = mainObject;
	}

	/**
	 * Gibt die Simulationsvariante zurück
	 *
	 * @return Simulationsvariante
	 */
	public short getSimVar() {
		return _simVar;
	}

	/**
	 * Setzt die Simulationsvariante
	 *
	 * @param simVar Simulationsvariante
	 *
	 * @throws FailureException Fehlerhafter Parameter
	 */
	public void setSimVar(final short simVar) throws FailureException {
		if(simVar < 0 || simVar > 999) {
			throw new FailureException(InterpreterErrorMessage.INVALID_SIMVAR + simVar, LoggerException.OTHER);
		}
		this._simVar = simVar;
	}

	/**
	 * Bindungen der freien Aspekte (optional)
	 *
	 * @return Map
	 */
	public Map<String, String> getAspBindings() {
		return Collections.unmodifiableMap(_aspBindings);
	}

	public void setAspBindings(final Map<String, String> aspBindings) throws FailureException {
		if(aspBindings == null) throw new FailureException(ErrorMessage.PARAM_IS_NULL + "aspBindings", LoggerException.OTHER);
		this._aspBindings = new HashMap<>(aspBindings);
	}

	/**
	 * Pseudoobjekt-Definitionen (optional). &lt;x,y&gt; x: Alias, y die dem Alias x zugeordnete Liste mit Objekt-Pids oder Leerstrings
	 *
	 * @return Pseudoobjekt-Definitionen
	 */
	public Map<String, String[]> getPseudoObjects() {
		final Map<String, String[]> tempMap = new HashMap<>(_pseudoObjects.size());
		for(final Map.Entry<String, List<String>> entry : _pseudoObjects.entrySet()) {
			tempMap.put(entry.getKey(), entry.getValue().toArray(new String[entry.getValue().size()]));
		}
		return tempMap;
	}

	/**
	 * Pseudoobjekt-Definitionen (optional). &lt;x,y&gt; x: Alias, y die dem Alias x zugeordnete Liste mit Objekt-Pids oder Leerstrings
	 *
	 * @return Pseudoobjekt-Definitionen
	 */
	public Map<String, List<String>> getPseudoObjects2() {
		return Collections.unmodifiableMap(_pseudoObjects);
	}

	/**
	 * Setzt die Pseudoobjekt-Definitionen
	 *
	 * @param pseudoObjects Zuordnung Alias zu PseudoObjekt-Pids. Alle Listen müssen die gleiche Länge haben. Leerstrings oder null-Einträge sind in den Listen
	 *                      gültig, für diese Einträge wird eine leere Platzhalterspalte erstellt. Die Klasse speichert eine Kopie des pseudoObjects-Parameters.
	 *
	 * @throws FailureException Bei ungültigen Parametern
	 */
	public void setPseudoObjects2(final Map<String, List<String>> pseudoObjects) throws FailureException {
		int oldLength = -1;
		if(pseudoObjects == null) throw new FailureException(ErrorMessage.PARAM_IS_NULL + "pseudoObjects", LoggerException.OTHER);
		this._pseudoObjects = new HashMap<>(pseudoObjects);
		for(final Map.Entry<String, List<String>> pseudoObjectEntry : _pseudoObjects.entrySet()) {
			final int length = pseudoObjectEntry.getValue().size();
			if(pseudoObjectEntry.getKey() == null) throw new FailureException(ErrorMessage.PARAM_IS_NULL + "pseudoObjects.key", LoggerException.OTHER);
			if(oldLength != -1 && oldLength != length) {
				throw new FailureException(ErrorMessage.INVALID_PARAM_VALUE + "Unterschiedliche Arraylänge in pseudoObjects", LoggerException.OTHER);
			}
			oldLength = length;
			// Liste klonen und dabei null-Werte ersetzen
			final List<String> listCopy = new ArrayList<>(pseudoObjectEntry.getValue());
			for(int i = 0; i < length; i++) {
				if(listCopy.get(i) == null) listCopy.set(i, "");
			}
			pseudoObjectEntry.setValue(Collections.unmodifiableList(listCopy));
		}
	}

	/**
	 * Setzt die Pseudoobjekt-Definitionen
	 *
	 * @param pseudoObjects Zuordnung Alias zu PseudoObjekt-Pids. Alle Arrays müssen die gleiche Länge haben. Leerstrings oder null-Einträge sind in den Arrays
	 *                      gültig, für diese Einträge wird eine leere Platzhalterspalte erstellt. Die Klasse speichert eine Kopie des pseudoObjects-Parameters.
	 *
	 * @throws FailureException Bei ungültigen Parametern
	 */
	public void setPseudoObjects(final Map<String, String[]> pseudoObjects) throws FailureException {
		final Map<String, List<String>> tempMap = new HashMap<>(pseudoObjects.size());
		for(final Map.Entry<String, String[]> entry : pseudoObjects.entrySet()) {
			tempMap.put(entry.getKey(), Arrays.asList(entry.getValue()));
		}
		setPseudoObjects2(tempMap);
	}

	/**
	 * Gibt zurück ob das protokoll gespeichert werden soll.
	 *
	 * @return true wenn das Protokoll gespeichert werden soll
	 */
	public boolean getSaveProtocol() {
		return _saveProtocol;
	}

	/**
	 * Setzt ob das Protokoll gespeichert werden soll.
	 *
	 * @param saveProtocol true wenn das Protokoll gespeichert werden soll
	 */
	public void setSaveProtocol(final boolean saveProtocol) {
		this._saveProtocol = saveProtocol;
	}

	/**
	 * Gibt die Protokollart zurück.
	 * @return Protokollart
	 * @see #setProtocolType(ProtocolType)
	 */
	public ProtocolType getProtocolType() {
		return _protocolType;
	}

	/**
	 * Setzt die Protokollart.
	 * @param protocolType Zu erstellende Protokollart. <code>null</code> wird als {@link de.bsvrz.sys.funclib.losb.datk.ProtocolType#Undefined} interpretiert
	 *                     und bedeutet das die im Skript definierte Protokollart verwendet werden soll. Wenn das Skript keine Vorgabe macht, wird ein
	 *                     Zustandsprotokoll erzeugt.
	 */
	public void setProtocolType(ProtocolType protocolType) {
		if(protocolType == null) protocolType = ProtocolType.Undefined;
		_protocolType = protocolType;
	}

	/**
	 * Gibt die Markierungsart für "Keine Änderungen" zurück.
	 * @return Markierungsart, entweder Zeilenweise oder für jede einzelne Zelle/jeden einzelnen Datensatz.
	 * @see #setNoChangeMarker(NoChangeMarker)
	 */
	public NoChangeMarker getNoChangeMarker() {
		return _noChangeMarker;
	}

	/**
	 * Setzt die Markierungsart für "Keine Änderungen" bei Zustandsprotokollen. Zeilenweise NoChange-Markierungen werden durch einen
	 * {@link de.bsvrz.pua.prot.client.dataobject.ProtocolNoChanges}-Datensatz abgebildet. Bei Zellenweisen Markierungen werden die einzelnen Zellen
	 * mit einem ExpressionResult vom Typ {@link de.bsvrz.pua.prot.util.ExpressionResult.ResultType#NO_CHANGE} befüllt.
	 * @param noChangeMarker Markierungsart, entweder Zeilenweise oder für jede einzelne Zelle/jeden einzelnen Datensatz.
	 *                       <code>null</code> wird als {@link de.bsvrz.sys.funclib.losb.datk.NoChangeMarker#Undefined} interpretiert
	 *                       und bedeutet das die im Skript definierte Markierungsart verwendet werden soll. Wenn das Skript keine Vorgabe macht, werden
	 *                       zeilenweise Markierungen verwendet.
	 */
	public void setNoChangeMarker(NoChangeMarker noChangeMarker) {
		if(noChangeMarker == null) noChangeMarker = NoChangeMarker.Undefined;
		_noChangeMarker = noChangeMarker;
	}

	/**
	 * Gibt zurück, ob ein Änderungsprotokoll erstellt werden soll.
	 *
	 * @return true: Ja, false: Nein, null: Standardwerte im Skript verwenden (nein, falls nicht definiert).
	 *
	 * @deprecated {@link #getProtocolType()} benutzen, da hier auch weitere Protokolltypen wie Ereignisprotokoll unterstützt werden
	 */
	@Deprecated
	public Boolean getDeltaProtocol() {
		switch(_protocolType){
			case StatusProtocol:
				return false;
			case DeltaProtocol:
				return true;
			case EventProtocol:
				return false;
			case Undefined:
				return null;
		}
		throw new IllegalStateException("_protocolType darf nicht null sein");
	}

	/**
	 * Setzt, ob ein Änderungsprotokoll erstellt werden soll.
	 *
	 * @param deltaProtocol true: Ja, false: Nein, null: Standardwerte im Skript verwenden (nein, falls nicht definiert).
	 *
	 * @deprecated {@link #setProtocolType(ProtocolType)}} benutzen, da hier auch weitere Protokolltypen wie Ereignisprotokoll unterstützt werden
	 */
	@Deprecated
	public void setDeltaProtocol(final Boolean deltaProtocol) {
		if(deltaProtocol == null){
			setProtocolType(ProtocolType.Undefined);
			return;
		}
		if(deltaProtocol){
			setProtocolType(ProtocolType.DeltaProtocol);
			return;
		}
		setProtocolType(ProtocolType.StatusProtocol);
	}

	/**
	 * Art der Daten, die vom Archivsystem angefordert werden sollen
	 *
	 * @return ArchiveDataKind-Set
	 */
	public Set<ArchiveDataKind> getArchiveDataKind() {
		return Collections.unmodifiableSet(_archiveDataKind);
	}

	/**
	 * Art der Daten, die vom Archivsystem angefordert werden sollen
	 *
	 * @return ArchiveDataKindCombination
	 */
	public ArchiveDataKindCombination getArchiveDataKindCombination() {
		final ArrayList<ArchiveDataKind> dataKinds = new ArrayList<>(getArchiveDataKind());
		switch(dataKinds.size()) {
			case 1:
				return new ArchiveDataKindCombination(dataKinds.get(0));
			case 2:
				return new ArchiveDataKindCombination(dataKinds.get(0), dataKinds.get(1));
			case 3:
				return new ArchiveDataKindCombination(dataKinds.get(0), dataKinds.get(1), dataKinds.get(2));
			case 4:
				return new ArchiveDataKindCombination(dataKinds.get(0), dataKinds.get(1), dataKinds.get(2), dataKinds.get(3));
		}
		throw new IllegalStateException("Ungültige Datensatzart");
	}

	/**
	 * Setzt die Datensatzarten für die das Skript erstellt werden soll
	 *
	 * @param archiveDataKind Datensatzarten
	 *
	 * @throws FailureException Fehlerhafter Parameter
	 */
	public void setArchiveDataKind(final Collection<ArchiveDataKind> archiveDataKind) throws FailureException {
		if(archiveDataKind == null) throw new FailureException(ErrorMessage.PARAM_IS_NULL + "archiveDataKind", LoggerException.OTHER);
		if(archiveDataKind.size() == 0) throw new FailureException(ErrorMessage.INVALID_PARAM_VALUE + "archiveDataKind.size() ist 0", LoggerException.OTHER);
		this._archiveDataKind = new HashSet<>(archiveDataKind);
	}

	/**
	 * Setzt die Datensatzarten für die das Skript erstellt werden soll
	 *
	 * @param archiveDataKind Datensatzarten
	 *
	 * @throws FailureException Fehlerhafter Parameter
	 */
	public void setArchiveDataKind(final ArchiveDataKind[] archiveDataKind) throws FailureException {
		if(archiveDataKind == null) throw new FailureException(ErrorMessage.PARAM_IS_NULL + "archiveDataKind", LoggerException.OTHER);
		if(archiveDataKind.length == 0) throw new FailureException(ErrorMessage.INVALID_PARAM_VALUE + "archiveDataKind.length ist 0", LoggerException.OTHER);
		this._archiveDataKind = new HashSet<>(Arrays.asList(archiveDataKind));
	}

	/**
	 * Setzt die Datensatzarten für die das Skript erstellt werden soll
	 *
	 * @param archiveDataKind Datensatzarten
	 */
	public void setArchiveDataKindCombination(final ArchiveDataKindCombination archiveDataKind) {
		this._archiveDataKind = new HashSet<>();
		if(archiveDataKind.isOnline()) _archiveDataKind.add(ArchiveDataKind.ONLINE);
		if(archiveDataKind.isOnlineDelayed()) _archiveDataKind.add(ArchiveDataKind.ONLINE_DELAYED);
		if(archiveDataKind.isRequested()) _archiveDataKind.add(ArchiveDataKind.REQUESTED);
		if(archiveDataKind.isRequestedDelayed()) _archiveDataKind.add(ArchiveDataKind.REQUESTED_DELAYED);
	}

	/**
	 * Text, der dem Protokoll mitgegeben werden kann und später in {@link de.bsvrz.pua.prot.client.dataobject.ProtocolAdministrationData#infoText} ausgelesen
	 * werden kann
	 *
	 * @return InfoText
	 */
	public String getInfoText() {
		return _infoText;
	}

	/**
	 * Setzt einen beschreibenden Freitext des Protokolls. Der Text wird mit dem Protokoll in {@link de.bsvrz.pua.prot.client.dataobject.ProtocolAdministrationData#infoText}
	 * wieder zurückgegeben.
	 *
	 * @param infoText Freitext
	 *
	 * @throws FailureException null-Parameter
	 */
	public void setInfoText(final String infoText) throws FailureException {
		if(infoText == null) throw new FailureException(ErrorMessage.PARAM_IS_NULL + "infoText", LoggerException.OTHER);
		this._infoText = infoText;
	}

	/**
	 * Gibt den Erstellernamen des Protokolls zurück (Freitext)
	 *
	 * @return Name des Protokollerstellers
	 */
	public String getCreatorName() {
		return _creatorName;
	}

	/**
	 * Setzt den Namen des Protokollerstellers (Freitext)
	 *
	 * @param creatorName namen
	 *
	 * @throws FailureException null-Parameter
	 */
	public void setCreatorName(final String creatorName) throws FailureException {
		if(creatorName == null) throw new FailureException(ErrorMessage.PARAM_IS_NULL + "creatorName", LoggerException.OTHER);
		this._creatorName = creatorName;
	}

	/**
	 * Gibt den Benutzer zurück, der das Protokoll erstellt hat. Der Wert ist noch nicht initialisiert ({@code null}) wenn dieses Objekt noch
	 * benutzt wird um einen Protokollauftrag zu senden. Der Wert wird erst gefüllt, wenn PuA-Seitig das Protokoll gestartet wird.
	 * @return Benutzerobjekt
	 */
	@Nullable
	public DynamicObject getCreatorUser() {
		return _creatorUser;
	}

	/**
	 * Gibt das Erstellungsdatums des Protokolls zurück
	 *
	 * @return Erstellungsdatum (ms seit 1. jan 1970)
	 */
	public long getCreationDate() {
		return _creationDate;
	}

	/**
	 * Setzt das Erstellungsdatum des Protokolls
	 *
	 * @param creationDate Erstellungsdatum (ms seit 1. jan 1970)
	 */
	public void setCreationDate(final long creationDate) {
		this._creationDate = creationDate;
	}

	/**
	 * Ist das Protokoll öffentlich?
	 *
	 * @return Ist das Protokoll öffentlich
	 */
	public boolean isPublic() {
		return _isPublic;
	}

	/**
	 * Setzt, ob das Protokoll öffentlich sein soll
	 *
	 * @param isPublic true wenn öffentlich
	 */
	public void setIsPublic(final boolean isPublic) {
		_isPublic = isPublic;
	}

	/**
	 * PuA-interne Funktion, die das Benutzerobjekt aktualisiert, wenn ein Protokollauftrag empfangen wurde.
	 * @param processingParameter Objekt, das aktualisiert werden soll
	 * @param client Applikation
	 */
	public static void updateClient(final ProcessingParameter processingParameter, final SystemObject client) {
		if(client instanceof ClientApplication) {
			ClientApplication clientApplication = (ClientApplication) client;
			processingParameter._creatorUser = clientApplication.getUser();
		}
	}
}
