/*
 * Copyright 2005-2008 by beck et al. projects GmbH, Munich
 * Copyright 2009 by Kappich Systemberatung, Aachen
 *
 * This file is part of de.bsvrz.pua.prot.
 *
 * de.bsvrz.pua.prot is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * de.bsvrz.pua.prot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.pua.prot.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Contact Information:
 * Kappich Systemberatung
 * Martin-Luther-Straße 14
 * 52062 Aachen, Germany
 * phone: +49 241 4090 436
 * mail: <info@kappich.de>
 */

package de.bsvrz.pua.prot.util.attributes;

import antlr.collections.AST;
import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.pua.prot.aggregations.Aggregation;
import de.bsvrz.pua.prot.processing.util.BaseDataSet;
import de.bsvrz.pua.prot.processing.util.ExpressionResultAndState;
import de.bsvrz.pua.prot.processing.util.TempAttribut;
import de.bsvrz.pua.prot.util.ExpressionTree;
import de.bsvrz.pua.prot.util.Qualifier;

import java.util.*;

/**
 * Repräsentiert ein temporäres Attribut.
 *
 * @author beck et al. projects GmbH
 * @author Martin Hilgers
 * @version $Revision$ / $Date$ / ($Author$)
 */
public class TempAttributeDescription extends AttributeDescription implements Cloneable{

	/** Zugeordneter Ausdruck */
	private ExpressionTree _expression;

	private SystemObject _type;

	private AST _AST;

	private int _index = -1;


	/** Konstruktor. Initialisiert {@link #_expression} und {@link #_type} mit <code>null</code>. */
	public TempAttributeDescription() {
		_expression = null;
		_type = null;
	}

	/**
	 * @param expression  Zugeordneter Ausdruck.
	 * @param type        Typ des temporären Attributs.
	 * @param columnName  Spaltenname. Kann <code>null</code> sein.
	 * @param aggregations Zu verwendende Aggregation. Kann <code>null</code> sein.
	 * @param name        Name des Attributs. Kann <code>null</code> sein.
	 */
	public TempAttributeDescription(final ExpressionTree expression, final SystemObject type, final String columnName, final List<Class<? extends Aggregation>> aggregations, final String name) {
		super(columnName, aggregations, name);
		this._expression = expression;
		this._type = type;
	}

	/**
	 * Klont die übergebene TempAttributeDescription und verwendet den übergebenen Spaltennamen
	 * @param tempAttributeDescription Original-Objekt
	 * @param columnName Spaltenname
	 */
	public TempAttributeDescription(final TempAttributeDescription tempAttributeDescription, final String columnName) {
		
		_expression= tempAttributeDescription.getExpression();
		setName(tempAttributeDescription.getName());
		_AST = (tempAttributeDescription.getAST());
		_index = tempAttributeDescription._index;
		_type = tempAttributeDescription._type;
		setAggregationClasses(tempAttributeDescription.getAggregationClasses());
		setColumnName(columnName);
	}

	public TempAttributeDescription(final String attribute, final SystemObject objectType, final AST ast, final ExpressionTree expression) {
		setName(attribute);
		_type = objectType;
		_AST  = ast;
		_expression = expression;
	}

	public TempAttributeDescription(final String name, final String columnName) {
		setName(name);
		setColumnName(columnName);
	}


	/**
	 * Gibt die verwendeten Dataindices zurück. Falls das temporäre Attribut selbst im Ausdruck vorkommt (Rekursion) wird es nicht weiter durchsucht.
	 *
	 * @return Datenindices des AttributeDescription
	 *
	 * @see de.bsvrz.pua.prot.util.attributes.Elements#getQualifiers()
	 */
	@Override
    public Collection<Qualifier> getQualifiers() {
		final HashSet<Qualifier> h = new HashSet<>();

		if(_expression != null) {
			for(final AttributeDescription att : _expression.getAttributes()) {
				if(!this.equals(att)) h.addAll(att.getQualifiers());
			}
		}
		return h;
	}

	/**
	 * Der Hashwert des Objekts ist nur abhängig vom Namen.
	 *
	 * @return Hashwert des Objekts
	 *
	 * @see java.lang.Object#hashCode()
	 * @see AttributeDescription#_name
	 */
	@Override
    public int hashCode() {
		int result = 0;

		if(getName() != null) result = getName().hashCode();

		return result;
	}

	/** @see java.lang.Object#toString() */
	@Override
    public String toString() {
		String result = "Temporäres Attribut " + getName() + " ";
		try {
			result += _type.getType();
		}
		catch(Exception ignored) {
		}

		if(getColumnName() != null) result += " als \"" + getColumnName() + "\"";
		if(getAggregationClasses() != null) result += ", aggregiert durch " + getAggregationClasses().toString();
		result += " :: " + _expression;
		result += " [" + _index + "]";
		return result;
	}

	/**
	 * Prüft auf Gleichheit von {@link AttributeDescription#_name}.
	 *
	 * @see java.lang.Object#equals(java.lang.Object)
	 */
	@Override
    public boolean equals(final Object obj) {
		if(obj instanceof TempAttributeDescription) {
			final TempAttributeDescription other = (TempAttributeDescription)obj;
			return Objects.equals(getName(), other.getName()) && Objects.equals(_index, other._index);
		}
		else {
			return false;
		}
	}

	/**
	 * Liefert ein Set mit den temporären Attributen, die dieses temporäre Attribut direkt verwendet. Dazu wird der Ausdruck, der diesem temporären Attribut
	 * zugeordnet ist, durchsucht, und die gefundenen temporären Attribute zurückgegeben.
	 *
	 * @return Temporäre Attribute. Set kann null sein, wenn dem temporären Attribut noch kein Ausdruck zugeordnet wurde.
	 */
	public HashSet<TempAttributeDescription> getTempAttributes() {
		final HashSet<TempAttributeDescription> result;
		if(_expression != null) {
			result = _expression.getDirectTempAttributes();
		}
		else {
			result = new HashSet<>();
		}
		return result;
	}


	/**
	 * Liefert die Menge der referenzierten realen Attribute zurück. Es werden nur direkt referenzierte reale Attribute zurückgegeben. D.h. temporäre Attribute und
	 * deren reale Attribute werden nicht zurückgegeben.
	 *
	 * @see de.bsvrz.pua.prot.util.attributes.Elements#getRealElements()
	 */
	@Override
	public HashSet<RealAttributeDescription> getRealElements() {
		if(_expression != null) {
			return _expression.getRealAttributes();
		}
		else {
			return new HashSet<>();
		}
	}

	/**
	 * Dem temporären Attribut zugeordneter Ausdruck
	 *
	 * @return Zugeordneter Ausdruck.
	 */
	public ExpressionTree getExpression() {
		return _expression;
	}

	/**
	 * Setzt den Ausdruck, durch den der Wert des Attributs bestimmt wird.
	 *
	 * @param expression Ausdruck.
	 */
	public void setExpression(final ExpressionTree expression) {
		this._expression = expression;
	}

	/**
	 * Liefert den Wert des temporären Attributs zurück.
	 *
	 * @param realElements Reale Attribute
	 * @param tempElements Temporäre Attribute
	 *
	 * @param pseudoObjectIndex PseudoObjectIndex
	 * @return Wert des temporären Attributs. <code>null</code> im Fehlerfall.
	 */
	public ExpressionResultAndState getValue(
			final Hashtable<RealElement, BaseDataSet> realElements, final Hashtable<TempAttributeDescription, TempAttribut> tempElements, final int pseudoObjectIndex) {
		//OPT Sollte nicht jedes Mal neu berechnet werden.
		if(_expression == null) {
			return null;
		}
		else {
			return _expression.evaluateEx(realElements, tempElements, pseudoObjectIndex);
		}
	}

	/**
	 * Liefert den Wert des temporären Attributs zurück.
	 *
	 * @param realElements Reale Attribute
	 * @param tempElements Temporäre Attribute
	 *
	 * @return Wert des temporären Attributs. <code>null</code> im Fehlerfall.
	 */
	public ExpressionResultAndState getValue(
			final Hashtable<RealElement, BaseDataSet> realElements, final Hashtable<TempAttributeDescription, TempAttribut> tempElements) {
		return getValue(realElements, tempElements, _index);
	}

	/**
	 * Setzt den AST für dieses Element (damit Fehlermeldungen usw. die Zeilennummer erhalten können)
	 * @param AST AST
	 */
	public void setAST(final AST AST) {
		_AST = AST;
	}

	/**
	 * Gibt den gesetzten AST zurück
	 * @return AST
	 */
	public AST getAST() {
		return _AST;
	}

	/**
	 * Typ des temporären Attributs
	 * @return Typ
	 */
	public SystemObject getType() {
		return _type;
	}

	public void setType(final SystemObject type) {
		_type = type;
	}

	public TempAttributeDescription createIndexedClone(final int index) {
		final TempAttributeDescription clone;
		try {
			clone = (TempAttributeDescription)super.clone();
			clone.setIndex(index);
		}
		catch(CloneNotSupportedException ignored) {
			throw new IllegalStateException("Konnte Objekt nicht klonen");
		}
		return clone;
	}

	/**
	 * Setzt den Alias-Index
	 * @param index Alias-Index
	 */
	public void setIndex(final int index) {
		_index = index;
	}

	/**
	 * Gibt den Index des verwendeten Alias an.
	 * @return Index des verwendeten Alias-Segments oder -1 falls keine Aliase verwendet werden.
	 */
	public int getIndex() {
		return _index;
	}
}
